from _decimal import Decimal
from fractions import Fraction
from pathlib import Path, PosixPath

import pytest

from katalytic.data.checks import (
    is_generator_function, is_iterator, dicts_share_key_order, dicts_share_value_order, is_equal
)
# noinspection PyProtectedMember
from katalytic.data import (
    _C, _C_obj, _callables, _collections, _dict_views, _flatten, _function, _functions, _generators, _iterables,
    _iterators, _lambda, _numbers, _obj, _objects, _primitives, _sequences, _singletons, _strings, _types, all_types,
    all_types_besides, as_list_of_dicts, as_list_of_lists, first, first_with_idx, flatten, flatten_recursive, last,
    last_with_idx, map_dict_keys, map_dict_values, map_recursive, one, pick_all, pick_all_besides, pick_any, pop_max,
    pop_max_by_dict_key, pop_max_by_dict_value, pop_min, pop_min_by_dict_key, pop_min_by_dict_value, sort_dict_by_keys,
    sort_dict_by_keys_recursive, sort_dict_by_values, sort_dict_by_values_recursive, as_dict_of_lists, sort_recursive,
    swap_keys_and_values, xor, xor_with_idx, detect_fronts, detect_fronts_positive, detect_fronts_negative
)


def _is_list(x):
    return isinstance(x, list)


def _is_negative(x):
    return _is_num(x) and x < 0


def _is_num(x):
    return isinstance(x, (int, float))


def _is_odd(x):
    return _is_num(x) and x % 2 == 1


def _is_str(x):
    return isinstance(x, str)


class Test_all_types:
    @pytest.mark.parametrize('wrong_type', [1, 1.0, True, False, object()])
    def test_wrong_type(self, wrong_type):
        with pytest.raises(TypeError):
            all_types(wrong_type)

    @pytest.mark.parametrize('unexpected', [
        ['iterable', 'func', 'strong']
    ])
    def test_unexpected(self, unexpected):
        with pytest.raises(ValueError):
            all_types(unexpected)

    @pytest.mark.parametrize('whitelist, expected', [
        ('callables', _callables),
        ('collections', _collections),
        ('dict_views', _dict_views),
        ('functions', _functions),
        ('generators', _generators),
        ('iterators', _iterators),
        ('numbers', _numbers),
        ('objects', _objects),
        ('primitives', _primitives),
        ('sequences', _sequences),
        ('singletons', _singletons),
        ('strings', _strings),
        (None, _flatten(_types.values())),
        (['iterables', 'objects', 'path'], [*_iterables, *_objects, Path('')]),
        (['iterables'], _iterables),
    ])
    def test_all_types(self, whitelist, expected):
        assert all_types(whitelist) == expected


class Test_all_types_besides:
    @pytest.mark.parametrize('wrong_type', [1, 1.0, True, False, None, object()])
    def test_wrong_type(self, wrong_type):
        with pytest.raises(TypeError):
            all_types_besides(wrong_type)

    @pytest.mark.parametrize('unexpected', [
        ['iterable', 'func', 'strong']
    ])
    def test_unexpected(self, unexpected):
        with pytest.raises(ValueError):
            all_types_besides(unexpected)

    @pytest.mark.parametrize('blacklist, expected', [
        (
            ['iterables'],
            [
                True, False, bytearray(b''), b'', _function, _lambda, _C_obj, _C, 0j, Decimal('0'), 0.0, Fraction(0, 1),
                0, None, _obj, PosixPath('.'), ''
            ]
        ),
        (
            'iterables',
            [
                True, False, bytearray(b''), b'', _function, _lambda, _C_obj, _C, 0j, Decimal('0'), 0.0, Fraction(0, 1),
                0, None, _obj, PosixPath('.'), ''
            ]
        ),
        (
            ['iterables', 'generators', 'functions', 'objects', 'path'],
            [True, False, bytearray(b''), b'', _C, 0j, Decimal('0'), 0.0, Fraction(0, 1), 0, None, '']
        ),
    ])
    def test_all_types_besides(self, blacklist, expected):
        assert all_types_besides(blacklist) == expected


# noinspection PyTypeChecker
class Test_as_dict_of_lists:
    @pytest.mark.parametrize('mistake', all_types_besides('booleans'))
    def test_precondition_empty_ok(self, mistake):
        with pytest.raises(TypeError):
            as_dict_of_lists([], empty_ok=mistake)

    @pytest.mark.parametrize('wrong_type', [1, 1.0, True, False, None, object(), iter([]), map(lambda x: x, [])])
    def test_wrong_type(self, wrong_type):
        with pytest.raises(TypeError):
            as_dict_of_lists(wrong_type)

    @pytest.mark.parametrize('data, expected', [
        # from dict of sequences
        [
            {'b': [3, 4], 'a': [1, 2]},
            {'b': [3, 4], 'a': [1, 2]},
        ],
        [
            {'a': (1, 2), 'b': (3, 4)},
            {'a': [1, 2], 'b': [3, 4]},
        ],
        # from sequence of dicts
        [
            ({'a': 1, 'b': 2}, {'b': 4, 'a': 3}),
            {'a': [1, 3], 'b': [2, 4]},
        ],
        [
            [{'b': 2, 'a': 1}, {'b': 4, 'a': 3}],
            {'b': [2, 4], 'a': [1, 3]},
        ],
        # from sequence of sequences
        [
            [['b', 'a'], [1, 2], [3, 4]],
            {'b': [1, 3], 'a': [2, 4]},
        ],
        [
            (('a', 'b'), (1, 2), (3, 4)),
            {'a': [1, 3], 'b': [2, 4]},
        ]
    ])
    def test_preserves_key_order_and_converts_seq_to_list(self, data, expected):
        actual = as_dict_of_lists(data)
        assert actual == expected
        # noinspection PyUnresolvedReferences
        assert list(actual.keys()) == list(expected.keys())

    @pytest.mark.parametrize('data', [[], (), {}])
    def test_empty_ok(self, data):
        assert as_dict_of_lists(data, empty_ok=True) == []

    @pytest.mark.parametrize('data', [[], (), {}])
    def test_empty_ok_False(self, data):
        with pytest.raises(ValueError):
            as_dict_of_lists(data, empty_ok=False)


# noinspection PyTypeChecker
class Test_as_list_of_dicts:
    @pytest.mark.parametrize('mistake', all_types_besides('booleans'))
    def test_precondition_empty_ok(self, mistake):
        with pytest.raises(TypeError):
            as_list_of_dicts([], empty_ok=mistake)

    @pytest.mark.parametrize('wrong_type', [1, 1.0, True, False, None, object(), iter([]), map(lambda x: x, [])])
    def test_wrong_type(self, wrong_type):
        with pytest.raises(TypeError):
            as_list_of_dicts(wrong_type)

    def test_different_keys(self):
        data = [{'a': 1, 'b': 2}, {'c': 3, 'd': 4, 'e': 5}]
        with pytest.raises(TypeError):
            as_list_of_dicts(data)

    @pytest.mark.parametrize('data, expected', [
        # from dict of sequences
        [
            {'a': [1, 2], 'b': [3, 4]},
            [{'a': 1, 'b': 3}, {'a': 2, 'b': 4}],
        ],
        [
            {'b': (3, 4), 'a': (1, 2)},
            [{'b': 3, 'a': 1}, {'b': 4, 'a': 2}],
        ],
        # from sequence of dicts
        [
            [{'b': 2, 'a': 1}, {'a': 3, 'b': 4}],
            [{'b': 2, 'a': 1}, {'b': 4, 'a': 3}],
        ],
        [
            ({'a': 1, 'b': 2}, {'a': 3, 'b': 4}),
            [{'a': 1, 'b': 2}, {'a': 3, 'b': 4}],
        ],
        # from sequence of sequences
        [
            [['b', 'a'], [1, 2], [3, 4]],
            [{'b': 1, 'a': 2}, {'b': 3, 'a': 4}],
        ],
        [
            (['a', 'b'], [1, 2], [3, 4]),
            [{'a': 1, 'b': 2}, {'a': 3, 'b': 4}],
        ],
    ])
    def test_preserves_key_order_and_converts_seq_to_list(self, data, expected):
        actual = as_list_of_dicts(data)
        keys = list(expected[0].keys())
        assert actual == expected
        assert all(list(d.keys()) == keys for d in actual)

    @pytest.mark.parametrize('data', [[], (), {}])
    def test_empty_ok(self, data):
        assert as_list_of_dicts(data, empty_ok=True) == []

    @pytest.mark.parametrize('data', [[], (), {}])
    def test_empty_ok_False(self, data):
        with pytest.raises(ValueError):
            as_list_of_dicts(data, empty_ok=False)


# noinspection PyTypeChecker
class Test_as_list_of_lists:
    @pytest.mark.parametrize('mistake', all_types_besides('booleans'))
    def test_precondition_empty_ok(self, mistake):
        with pytest.raises(TypeError):
            as_list_of_lists([], empty_ok=mistake)

    @pytest.mark.parametrize('wrong_type', [1, 1.0, True, False, None, object(), iter([]), map(lambda x: x, [])])
    def test_wrong_type(self, wrong_type):
        with pytest.raises(TypeError):
            as_list_of_lists(wrong_type)

    @pytest.mark.parametrize('data, expected', [
        # from dict of sequences
        [
            {'a': [1, 2], 'b': [3, 4]},
            [['a', 'b'], [1, 3], [2, 4]],
        ],
        [
            {'b': (3, 4), 'a': (1, 2)},
            [['b', 'a'], [3, 1], [4, 2]],
        ],
        # from sequence of dicts
        [
            ({'a': 1, 'b': 2}, {'a': 3, 'b': 4}),
            [['a', 'b'], [1, 2], [3, 4]],
        ],
        [
            [{'b': 2, 'a': 1}, {'a': 3, 'b': 4}],
            [['b', 'a'], [2, 1], [4, 3]],
        ],
        # from sequence of sequences
        [
            [['b', 'a'], [1, 2], [3, 4]],
            [['b', 'a'], [1, 2], [3, 4]],
        ],
        [
            (['a', 'b'], (1, 2), [3, 4]),
            [['a', 'b'], [1, 2], [3, 4]],
        ],
    ])
    def test_preserves_key_order_and_converts_seq_to_list(self, data, expected):
        assert as_list_of_lists(data) == expected

    @pytest.mark.parametrize('data', [[], (), {}])
    def test_empty_ok(self, data):
        assert as_list_of_lists(data, empty_ok=True) == []

    @pytest.mark.parametrize('data', [[], (), {}])
    def test_empty_ok_False(self, data):
        with pytest.raises(ValueError):
            as_list_of_lists(data, empty_ok=False)


class Test_first:
    @pytest.mark.parametrize('wrong_type', [
        iter([]), map(lambda x: x, []), range(1), object(), [], {}, (), set(), 1, True, False, None, ''
    ])
    def test_key_not_callable(self, wrong_type):
        with pytest.raises(TypeError):
            first([], condition=wrong_type)

    def test_sets_are_not_allowed(self):
        with pytest.raises(TypeError):
            first(set())

    @pytest.mark.parametrize('data', [iter([]), map(lambda x: x, []), [], {}, ()])
    def test_empty(self, data):
        assert first(data) is None

    def test_no_first_with_condition(self):
        data = range(10)
        assert first(data, condition=lambda x: x > 50) is None

    @pytest.mark.parametrize('data, expected', [
        (range(10), 0),
        ([False, None, 5, True], False),
        (['', 0, None, False], ''),
        (map(lambda x: x+1, range(0, 100, 10)), 1),
    ])
    def test_first(self, data, expected):
        assert first(data) == expected


class Test_first_with_idx:
    @pytest.mark.parametrize('wrong_type', [
        iter([]), map(lambda x: x, []), range(1), object(), [], {}, (), set(), 1, True, False, None, ''
    ])
    def test_key_not_callable(self, wrong_type):
        with pytest.raises(TypeError):
            first_with_idx([], condition=wrong_type)

    def test_sets_are_not_allowed(self):
        with pytest.raises(TypeError):
            first_with_idx(set())

    @pytest.mark.parametrize('data', [iter([]), map(lambda x: x, []), [], {}, ()])
    def test_empty(self, data):
        assert first_with_idx(data) is None

    def test_no_first_with_condition(self):
        data = range(10)
        assert first_with_idx(data, condition=lambda x: x > 50) is None

    def test_first_with_key(self):
        data = range(5, 15)
        assert first_with_idx(data, condition=lambda x: x >= 10) == (5, 10)

    @pytest.mark.parametrize('data, expected', [
        (range(10), (0, 0)),
        ([False, None, 5, True], (0, False)),
        (['', 0, None, False], (0, '')),
        (map(lambda x: x+1, range(0, 100, 10)), (0, 1)),
    ])
    def test_first(self, data, expected):
        assert first_with_idx(data) == expected


class Test_last:
    @pytest.mark.parametrize('wrong_type', [
        iter([]), map(lambda x: x, []), range(1), object(), [], {}, (), set(), 1, True, False, None, ''
    ])
    def test_key_not_callable(self, wrong_type):
        with pytest.raises(TypeError):
            last([], condition=wrong_type)

    def test_sets_are_not_allowed(self):
        with pytest.raises(TypeError):
            last(set())

    @pytest.mark.parametrize('data', [iter([]), map(lambda x: x, []), [], {}, ()])
    def test_empty(self, data):
        assert last(data) is None

    def test_no_last_with_condition(self):
        data = range(10)
        assert last(data, condition=lambda x: x > 50) is None

    @pytest.mark.parametrize('data, expected', [
        (range(10), 9),
        ([False, None, 5, True], True),
        ([False, None, True, 5], 5),
        (['', 0, None, False], False),
        (map(lambda x: x+1, range(0, 100, 10)), 91),
    ])
    def test_last(self, data, expected):
        assert last(data) == expected


class Test_last_with_idx:
    @pytest.mark.parametrize('wrong_type', [
        iter([]), map(lambda x: x, []), range(1), object(), [], {}, (), set(), 1, True, False, None, ''
    ])
    def test_key_not_callable(self, wrong_type):
        with pytest.raises(TypeError):
            last_with_idx([], condition=wrong_type)

    def test_sets_are_not_allowed(self):
        with pytest.raises(TypeError):
            last_with_idx(set())

    @pytest.mark.parametrize('data', [iter([]), map(lambda x: x, []), [], {}, ()])
    def test_empty(self, data):
        assert last_with_idx(data) is None

    def test_no_last_with_condition(self):
        data = range(10)
        assert last_with_idx(data, condition=lambda x: x > 50) is None

    def test_last_with_key(self):
        data = range(5, 15)
        assert last_with_idx(data, condition=lambda x: x <= 13) == (8, 13)

    @pytest.mark.parametrize('data, expected', [
        (range(7, 10), (2, 9)),
        ([False, None, 5, True], (3, True)),
        (['', 0, None, False], (3, False)),
        (map(lambda x: x+1, range(0, 100, 10)), (9, 91)),
    ])
    def test_last(self, data, expected):
        assert last_with_idx(data) == expected


class Test_flatten:
    @pytest.mark.parametrize('wrong_type', [1, 1.0, True, False, None, '', object()])
    def test_should_be_an_iterable(self, wrong_type):
        with pytest.raises(TypeError):
            flatten(wrong_type)

    @pytest.mark.parametrize('empty', [[], (), {}, set(), iter([])])
    def test_empty(self, empty):
        assert flatten(empty) == []

    @pytest.mark.parametrize('data, expected', [
        ([1, 2, 3], [1, 2, 3]),
        ((1, 2, 3), [1, 2, 3]),
        ({1, 2, 3}, [1, 2, 3]),
    ])
    def test_already_flat(self, data, expected):
        assert flatten(data) == expected

    @pytest.mark.parametrize('data, expected', [
        ((1, [(2, 3), (4, (5, 6))]), [1, (2, 3), (4, (5, 6))]),
        ([[1, (2, 3), (4, (5, 6))]], [1, (2, 3), (4, (5, 6))]),
        ({1, (2, 3), (4, (5, 6))}, [1, 2, 3, 4, (5, 6)]),
    ])
    def test_flattens_only_one_level(self, data, expected):
        actual = flatten(data)
        if isinstance(data, set):
            actual = set(actual)
            expected = set(expected)

        assert actual == expected


class Test_flatten_recursive:
    @pytest.mark.parametrize('wrong_type', [1, 1.0, True, False, None, '', object()])
    def test_should_be_an_iterable(self, wrong_type):
        with pytest.raises(TypeError):
            flatten_recursive(wrong_type)

    @pytest.mark.parametrize('empty', [[], (), {}, set(), iter([])])
    def test_empty(self, empty):
        assert flatten_recursive(empty) == []

    @pytest.mark.parametrize('data, expected', [
        ([1, 2, 3], [1, 2, 3]),
        ((1, 2, 3), [1, 2, 3]),
        ({1, 2, 3}, [1, 2, 3]),
    ])
    def test_already_flat(self, data, expected):
        assert flatten_recursive(data) == expected

    @pytest.mark.parametrize('data, expected', [
        ((1, [(2, 3), (4, (5, 6))]), [1, 2, 3, 4, 5, 6]),
        ([[1, (2, 3), (4, (5, 6))]], [1, 2, 3, 4, 5, 6]),
        ({1, (2, 3), (4, (5, 6))}, [1, 2, 3, 4, 5, 6]),
    ])
    def test_flattens_all_levels(self, data, expected):
        actual = flatten_recursive(data)
        if isinstance(data, set):
            actual = set(actual)
            expected = set(expected)

        assert actual == expected


# noinspection PyTypeChecker
class Test_map_dict_keys:
    @pytest.mark.parametrize('wrong_type', [[], set(), (), {}, 1, 1.0, True, None, False, 'string', object()])
    def test_mapping_is_not_a_function(self, wrong_type):
        with pytest.raises(TypeError):
            map_dict_keys(wrong_type, {})

    @pytest.mark.parametrize('wrong_type', [[], set(), (), 1, 1.0, True, None, False, 'string', object()])
    def test_not_a_dict(self, wrong_type):
        with pytest.raises(TypeError):
            map_dict_keys(lambda x: x, wrong_type)

    @pytest.mark.parametrize('wrong_type', [[], set(), (), {}, 1, 1.0, True, False, 'string', object()])
    def test_condition_is_not_a_function(self, wrong_type):
        with pytest.raises(TypeError):
            map_dict_keys(lambda x: x, {}, condition=wrong_type)

    def test_empty(self):
        assert map_dict_keys(lambda x: x, {}) == {}

    def test_simple_mapping(self):
        assert map_dict_keys(str.upper, {'a': 1, 'b': 2}) == {'A': 1, 'B': 2}

    def test_conditioned_mapping(self):
        data = {'a': 1, (0, 1): 2}
        expected = {'A': 1, (0, 1): 2}
        assert map_dict_keys(str.upper, data, condition=_is_str) == expected


# noinspection PyTypeChecker
class Test_map_dict_values:
    @pytest.mark.parametrize('wrong_type', [[], set(), (), {}, 1, 1.0, True, None, False, 'string', object()])
    def test_mapping_is_not_a_function(self, wrong_type):
        with pytest.raises(TypeError):
            map_dict_values(wrong_type, {})

    @pytest.mark.parametrize('wrong_type', [[], set(), (), 1, 1.0, True, None, False, 'string', object()])
    def test_not_a_dict(self, wrong_type):
        with pytest.raises(TypeError):
            map_dict_values(lambda x: x, wrong_type)

    @pytest.mark.parametrize('wrong_type', [[], set(), (), {}, 1, 1.0, True, False, 'string', object()])
    def test_condition_is_not_a_function(self, wrong_type):
        with pytest.raises(TypeError):
            map_dict_values(lambda x: x, {}, condition=wrong_type)

    def test_empty(self):
        assert map_dict_values(lambda x: x, {}) == {}

    def test_simple_mapping(self):
        assert map_dict_values(str, {'a': 1, 'b': 2}) == {'a': '1', 'b': '2'}

    def test_conditioned_mapping(self):
        data = {'a': -1, (0, 1): 1}
        expected = {'a': '-1', (0, 1): 1}
        assert map_dict_values(str, data, condition=_is_negative) == expected


class Test_map_recursive:
    @pytest.mark.xfail(reason='Not implemented')
    def test_condition_is_iterable(self):
        """
        1. Should run both the `if is_iterable()` and the `if condition(x)` branches.
        2. Add another param for which of the two branches to run first"""
        assert 0

    @pytest.mark.parametrize('wrong_type', [[], set(), (), {}, 1, 1.0, True, None, False, 'string', object()])
    def test_mapping_is_not_a_function(self, wrong_type):
        with pytest.raises(TypeError):
            map_recursive(wrong_type, {})

    @pytest.mark.parametrize('wrong_type', [1, 1.0, True, None, False, 'string', object()])
    def test_not_an_iterable(self, wrong_type):
        with pytest.raises(TypeError):
            map_recursive(lambda x: x, wrong_type)

    @pytest.mark.parametrize('wrong_type', [[], set(), (), {}, 1, 1.0, True, False, 'string', object()])
    def test_condition_is_not_a_function(self, wrong_type):
        with pytest.raises(TypeError):
            map_recursive(lambda x: x, {}, condition=wrong_type)

    @pytest.mark.parametrize('wrong_type', [[], set(), (), {}, 1, 1.0, None, 'string', object()])
    def test_on_dict_keys_is_not_a_function(self, wrong_type):
        with pytest.raises(TypeError):
            map_recursive(lambda x: x, {}, on_dict_keys=wrong_type)

    @pytest.mark.parametrize('data', [[], (), {}, set()])
    def test_empty(self, data):
        assert map_recursive(lambda x: x, data) == data

    def test_simple(self):
        assert map_recursive(lambda x: x + 1, [1, 2]) == [2, 3]
        assert map_recursive(lambda x: x + 1, (1, 2)) == (2, 3)
        assert map_recursive(lambda x: x + 1, {1, 2}) == {2, 3}
        assert map_recursive(lambda x: x + 1, {'a': 1, 'b': 2}) == {'a': 2, 'b': 3}

        actual = map_recursive(lambda x: x + 1, iter([1, 2]))
        assert is_iterator(actual)
        assert tuple(actual) == (2, 3)

    def test_deep(self):
        data = [(1, {'a': 1, 'b': {1, 2}})]
        expected = [(2, {'a': 2, 'b': {2, 3}})]
        assert map_recursive(lambda x: x + 1, data) == expected

    def test_deep_with_condition(self):
        data = [(1, 2, {'a': 1, 'b': 2, 'c': {1, (1, 2)}})]
        expected = [(2, 2, {'a': 2, 'b': 2, 'c': {2, (2, 2)}})]
        assert map_recursive(lambda x: x + 1, data, condition=_is_odd) == expected

    def test_deep_with_condition_2(self):
        data = [1, 2, (1, 2, {'a': 1, 'b': [1, 2]})]
        expected = (1, 2, (1, 2, {'a': 1, 'b': (1, 2)}))
        assert map_recursive(tuple, data, condition=_is_list) == expected

    def test_on_dict_keys(self):
        data = {'a': 1, 'b': {'a': 1, 'b': 2}}
        expected = {'A': 1, 'B': {'A': 1, 'B': 2}}
        assert map_recursive(lambda x: x.upper(), data, condition=_is_str, on_dict_keys=True) == expected

    def test_on_dict_keys_iter(self):
        data = {'a': 1, (1, 2): {'a': 1, (3, 4): 2}}
        expected = {'a': 2, (2, 3): {'a': 2, (4, 5): 3}}
        assert map_recursive(lambda x: x + 1, data, condition=_is_num, on_dict_keys=True) == expected


class Test_one:
    @pytest.mark.parametrize('wrong_type', [
        iter([]), map(lambda x: x, []), range(1), object(), [], {}, (), set(), 1, True, False, None, ''
    ])
    def test_key_not_callable(self, wrong_type):
        with pytest.raises(TypeError):
            one([], condition=wrong_type)

    @pytest.mark.parametrize('wrong_type', [object(), 1, True, False, None, ''])
    def test_key_not_iterable(self, wrong_type):
        with pytest.raises(TypeError):
            one(wrong_type)

    @pytest.mark.parametrize('data', [iter([]), map(lambda x: x, []), [], {}, (), set()])
    def test_empty(self, data):
        assert one(data) is None

    @pytest.mark.parametrize('data', [{'', 0, None, False}, ['', 0, None, False]])
    def test_no_one(self, data):
        assert one(data) == ''

    def test_no_one_with_condition(self):
        data = range(10)
        assert one(data, condition=lambda x: x > 50) is None

    @pytest.mark.parametrize('data, expected', [
        (range(10), 0),
        ([False, None, 5, True], False),
        ([False, None, True, 5], False),
        (map(lambda x: x+1, range(0, 100, 10)), 1),
    ])
    def test_one(self, data, expected):
        assert one(data) == expected

    @pytest.mark.parametrize('data', [
        {0, 1, 2, 3},
        {3, 2, 1, 0},
    ])
    def test_one_of_set(self, data):
        assert one(data) in data


class Test_pop_min:
    @pytest.mark.parametrize('wrong_type', all_types_besides('iterables'))
    def test_not_an_iterable(self, wrong_type):
        with pytest.raises(TypeError):
            pop_min(wrong_type)

    def test_dict(self):
        with pytest.raises(TypeError):
            pop_min({})

    @pytest.mark.parametrize('wrong_type', all_types_besides('callables'))
    def test_key_is_not_a_function(self, wrong_type):
        with pytest.raises(TypeError):
            pop_min([1], condition=wrong_type)

    @pytest.mark.parametrize('data', all_types('iterables'))
    def test_empty(self, data):
        if isinstance(data, dict):
            return
        elif is_generator_function(data):
            # noinspection PyCallingNonCallable
            data = data()
            next(data)

        with pytest.raises(ValueError):
            pop_min(data)

    @pytest.mark.parametrize('data, expected', [
        ([3,2,1,4], (1, [3,2,4])),
        ((0,-1,3,4), (-1, (0,3,4))),
        (map(float, (0,3,-4,5)), (-4, [0,3,5])),
    ])
    def test_no_key(self, data, expected):
        assert pop_min(data) == expected

    @pytest.mark.parametrize('data, expected', [
        ([3,2,1,4], (4, [3,2,1])),
        ((0,-1,3,4), (4, (0,-1,3))),
        (map(float, (0,3,-4,5)), (5, [0,3,-4])),
    ])
    def test_with_key(self, data, expected):
        assert pop_min(data, condition=lambda x: -x) == expected


class Test_pop_max:
    @pytest.mark.parametrize('wrong_type', all_types_besides('iterables'))
    def test_not_an_iterable(self, wrong_type):
        with pytest.raises(TypeError):
            pop_max(wrong_type)

    def test_dict(self):
        with pytest.raises(TypeError):
            pop_max({})

    @pytest.mark.parametrize('wrong_type', all_types_besides('callables'))
    def test_key_is_not_a_function(self, wrong_type):
        with pytest.raises(TypeError):
            pop_max([1], condition=wrong_type)

    @pytest.mark.parametrize('data', all_types('iterables'))
    def test_empty(self, data):
        if isinstance(data, dict):
            return
        elif is_generator_function(data):
            # noinspection PyCallingNonCallable
            data = data()
            next(data)

        with pytest.raises(ValueError):
            pop_max(data)

    @pytest.mark.parametrize('data, expected', [
        ([3,2,1,4], (4, [3,2,1])),
        ((0,-1,3,4), (4, (0,-1,3))),
        (map(float, (0,3,-4,5)), (5, [0,3,-4])),
    ])
    def test_no_key(self, data, expected):
        assert pop_max(data) == expected

    @pytest.mark.parametrize('data, expected', [
        ([3,2,1,4], (1, [3,2,4])),
        ({0,-1,3,4}, (-1, {0,3,4})),
        (map(float, (0,3,-4,5)), (-4, [0,3,5])),
    ])
    def test_with_key(self, data, expected):
        assert pop_max(data, condition=lambda x: -x) == expected


class Test_pop_max_by_dict_key:
    @pytest.mark.parametrize('wrong_type', all_types_besides('dict'))
    def test_not_a_dict(self, wrong_type):
        with pytest.raises(TypeError):
            pop_max_by_dict_key(wrong_type)

    @pytest.mark.parametrize('wrong_type', all_types_besides('callables'))
    def test_key_is_not_a_function(self, wrong_type):
        with pytest.raises(TypeError):
            pop_max_by_dict_key({'1': 1}, condition=wrong_type)

    def test_empty(self):
        with pytest.raises(ValueError):
            pop_max_by_dict_key({})

    @pytest.mark.parametrize('data, expected', [
        ({'a': 3, 'b': 2, 'c': 1, 'd': 4}, (('d', 4), {'a': 3, 'b': 2, 'c': 1})),
        ({'a': 0, 'b': -1, 'c': 4, 'd': 3}, (('d', 3), {'a': 0, 'b': -1, 'c': 4})),
    ])
    def test_no_key(self, data, expected):
        assert pop_max_by_dict_key(data) == expected

    @pytest.mark.parametrize('data, expected', [
        ({3: 'a', 2: 'b', 1: 'c', 4: 'd'}, ((1, 'c'), {3: 'a', 2: 'b', 4: 'd'})),
        ({0: 'a', -1: 'b', 4: 'c', 3: 'd'}, ((-1, 'b'), {0: 'a', 4: 'c', 3: 'd'})),
    ])
    def test_with_key(self, data, expected):
        assert pop_max_by_dict_key(data, condition=lambda x: -x) == expected


class Test_pop_min_by_dict_key:
    @pytest.mark.parametrize('wrong_type', all_types_besides('dict'))
    def test_not_a_dict(self, wrong_type):
        with pytest.raises(TypeError):
            pop_min_by_dict_key(wrong_type)

    @pytest.mark.parametrize('wrong_type', all_types_besides('callables'))
    def test_key_is_not_a_function(self, wrong_type):
        with pytest.raises(TypeError):
            pop_min_by_dict_key({'1': 1}, condition=wrong_type)

    def test_empty(self):
        with pytest.raises(ValueError):
            pop_min_by_dict_key({})

    @pytest.mark.parametrize('data, expected', [
        ({'a': 3, 'b': 2, 'c': 1, 'd': 4}, (('a', 3), {'b': 2, 'c': 1, 'd': 4})),
        ({'a': 0, 'b': -1, 'c': 4, 'd': 3}, (('a', 0), {'b': -1, 'c': 4, 'd': 3})),
    ])
    def test_no_key(self, data, expected):
        assert pop_min_by_dict_key(data) == expected

    @pytest.mark.parametrize('data, expected', [
        ({3: 'a', 2: 'b', 1: 'c', 4: 'd'}, ((4, 'd'), {3: 'a', 2: 'b', 1: 'c'})),
        ({0: 'a', -1: 'b', 4: 'c', 3: 'd'}, ((4, 'c'), {0: 'a', -1: 'b', 3: 'd'})),
    ])
    def test_with_key(self, data, expected):
        assert pop_min_by_dict_key(data, condition=lambda x: -x) == expected


class Test_pop_max_by_dict_value:
    @pytest.mark.parametrize('wrong_type', all_types_besides('dict'))
    def test_not_a_dict(self, wrong_type):
        with pytest.raises(TypeError):
            pop_max_by_dict_value(wrong_type)

    @pytest.mark.parametrize('wrong_type', all_types_besides('callables'))
    def test_key_is_not_a_function(self, wrong_type):
        with pytest.raises(TypeError):
            pop_max_by_dict_value({'1': 1}, condition=wrong_type)

    def test_empty(self):
        with pytest.raises(ValueError):
            pop_max_by_dict_value({})

    @pytest.mark.parametrize('data, expected', [
        ({'a': 3, 'b': 2, 'c': 1, 'd': 4}, (('d', 4), {'a': 3, 'b': 2, 'c': 1})),
        ({'a': 0, 'b': -1, 'c': 4, 'd': 3}, (('c', 4), {'a': 0, 'b': -1, 'd': 3})),
    ])
    def test_no_key(self, data, expected):
        assert pop_max_by_dict_value(data) == expected

    @pytest.mark.parametrize('data, expected', [
        ({3: 'a', 2: 'b', 1: 'c', 4: 'd'}, ((3, 'a'), {2: 'b', 1: 'c', 4: 'd'})),
        ({0: 'a', -1: 'b', 4: 'c', 3: 'd'}, ((0, 'a'), {-1: 'b', 4: 'c', 3: 'd'})),
    ])
    def test_with_key(self, data, expected):
        assert pop_max_by_dict_value(data, condition=lambda x: -ord(x)) == expected


class Test_pop_min_by_dict_value:
    @pytest.mark.parametrize('wrong_type', all_types_besides('dict'))
    def test_not_a_dict(self, wrong_type):
        with pytest.raises(TypeError):
            pop_min_by_dict_value(wrong_type)

    @pytest.mark.parametrize('wrong_type', all_types_besides('callables'))
    def test_key_is_not_a_function(self, wrong_type):
        with pytest.raises(TypeError):
            pop_min_by_dict_value({'1': 1}, condition=wrong_type)

    def test_empty(self):
        with pytest.raises(ValueError):
            pop_min_by_dict_value({})

    @pytest.mark.parametrize('data, expected', [
        ({'a': 3, 'b': 2, 'c': 1, 'd': 4}, (('c', 1), {'a': 3, 'b': 2, 'd': 4})),
        ({'a': 0, 'b': -1, 'c': 4, 'd': 3}, (('b', -1), {'a': 0, 'c': 4, 'd': 3})),
    ])
    def test_no_key(self, data, expected):
        assert pop_min_by_dict_value(data) == expected

    @pytest.mark.parametrize('data, expected', [
        ({3: 'a', 2: 'b', 1: 'c', 4: 'd'}, ((4, 'd'), {3: 'a', 2: 'b', 1: 'c'})),
        ({0: 'a', -1: 'b', 4: 'c', 3: 'd'}, ((3, 'd'), {0: 'a', -1: 'b', 4: 'c'})),
    ])
    def test_with_key(self, data, expected):
        assert pop_min_by_dict_value(data, condition=lambda x: -ord(x)) == expected


class Test_pick_all:
    @pytest.mark.parametrize('wrong_type', [object(), 1, True, False, None, ''])
    def test_needles_should_be_iterable(self, wrong_type):
        with pytest.raises(TypeError):
            pick_all(wrong_type, [])

    @pytest.mark.parametrize('wrong_type', [object(), 1, True, False, None, ''])
    def test_haystack_should_be_iterable(self, wrong_type):
        with pytest.raises(TypeError):
            pick_all([], wrong_type)

    @pytest.mark.parametrize('needles, haystack, expected', [
        ([], [], []),
        ((), {}, []),
        ((), [1], []),
        ((1, ), [], []),
    ])
    def test_empty(self, needles, haystack, expected):
        assert pick_all(needles, haystack) == expected

    @pytest.mark.parametrize('needles, haystack, expected', [
        ([1, 2], [3, 4], []),
        ([0, True], (1, False), []),
    ])
    def test_no_match(self, needles, haystack, expected):
        assert pick_all(needles, haystack) == expected

    @pytest.mark.parametrize('needles, haystack, expected', [
        ([1, 2, 3], [2, 3, 4], [2, 3]),
        ([False, True, None], {True, None}, [True, None]),
    ])
    def test_match(self, needles, haystack, expected):
        assert pick_all(needles, haystack) == expected


class Test_pick_all_besides:
    @pytest.mark.parametrize('wrong_type', [object(), 1, True, False, None, ''])
    def test_needles_should_be_iterable(self, wrong_type):
        with pytest.raises(TypeError):
            pick_all_besides(wrong_type, [])

    @pytest.mark.parametrize('wrong_type', [object(), 1, True, False, None, ''])
    def test_haystack_should_be_iterable(self, wrong_type):
        with pytest.raises(TypeError):
            pick_all_besides([], wrong_type)

    @pytest.mark.parametrize('needles, haystack, expected', [
        ([], [], []),
        ((), {}, []),
    ])
    def test_empty(self, needles, haystack, expected):
        assert pick_all_besides(needles, haystack) == expected

    @pytest.mark.parametrize('needles, haystack, expected', [
        ([1, 2, 3], [3, 4, 5], [1, 2]),
        ([0, True, None], {1, False, None}, [0, True]),
    ])
    def test_match(self, needles, haystack, expected):
        assert pick_all_besides(needles, haystack) == expected

    @pytest.mark.parametrize('needles, haystack, expected', [
        ([1, 2, 3], [1, 2, 3], []),
        ([False, True, None], {False, True, None}, []),
    ])
    def test_no_match(self, needles, haystack, expected):
        assert pick_all_besides(needles, haystack) == expected


class Test_pick_any:
    @pytest.mark.parametrize('wrong_type', [object(), 1, True, False, None, ''])
    def test_needles_should_be_iterable(self, wrong_type):
        with pytest.raises(TypeError):
            pick_any(wrong_type, [])

    @pytest.mark.parametrize('wrong_type', [object(), 1, True, False, None, ''])
    def test_haystack_should_be_iterable(self, wrong_type):
        with pytest.raises(TypeError):
            pick_any([], wrong_type)

    @pytest.mark.parametrize('needles, haystack, expected', [
        ([], [], None),
        ((), {}, None),
        ((), [1], None),
        ((1, ), [], None),
    ])
    def test_empty(self, needles, haystack, expected):
        assert pick_any(needles, haystack) == expected

    @pytest.mark.parametrize('needles, haystack, expected', [
        ([1, 2], [3, 4], None),
        ([0, True], {1, False}, None),
    ])
    def test_no_match(self, needles, haystack, expected):
        assert pick_any(needles, haystack) == expected

    @pytest.mark.parametrize('needles, haystack, expected', [
        ([1, 2, 3], [2, 3, 4], 2),
        ([False, True, None], {None, True}, True),
    ])
    def test_match(self, needles, haystack, expected):
        assert pick_any(needles, haystack) == expected


class Test_sort_dict_by_keys:
    @pytest.mark.parametrize('wrong_type', [[], set(), (), 1, 1.0, True, None, False, 'string', object()])
    def test_not_a_dict(self, wrong_type):
        with pytest.raises(TypeError):
            sort_dict_by_keys(wrong_type)

    @pytest.mark.parametrize('wrong_type', [[], set(), (), {}, 1, 1.0, True, False, 'string', object()])
    def test_key_is_not_a_function(self, wrong_type):
        with pytest.raises(TypeError):
            sort_dict_by_keys({}, condition=wrong_type)

    @pytest.mark.parametrize('wrong_type', [[], set(), (), {}, 1, 1.0, None, 'string', object()])
    def test_not_a_bool(self, wrong_type):
        with pytest.raises(TypeError):
            sort_dict_by_keys({}, reverse=wrong_type)

    def test_empty(self):
        assert sort_dict_by_keys({}) == {}

    def test_simple(self):
        actual = sort_dict_by_keys({'b': 2, 'c': 1, 'a': 3})
        expected = {'a': 3, 'b': 2, 'c': 1}
        assert actual == expected
        assert dicts_share_key_order(actual, expected)

    def test_with_key(self):
        actual = sort_dict_by_keys({'qwerty': 1, 'x': 3, 'asd': 2}, condition=len)
        expected = {'x': 3, 'asd': 2, 'qwerty': 1}
        assert actual == expected
        assert dicts_share_key_order(actual, expected)

    def test_reversed(self):
        actual = sort_dict_by_keys({'b': 2, 'c': 1, 'a': 3}, reverse=True)
        expected = {'c': 1, 'b': 2, 'a': 3}
        assert actual == expected
        assert dicts_share_key_order(actual, expected)

    def test_with_key_and_reversed(self):
        actual = sort_dict_by_keys({'qwerty': 1, 'x': 3, 'asd': 2}, condition=len, reverse=True)
        expected = {'qwerty': 1, 'asd': 2, 'x': 3}
        assert actual == expected
        assert dicts_share_key_order(actual, expected)


class Test_sort_dict_by_values:
    @pytest.mark.parametrize('wrong_type', [[], set(), (), 1, 1.0, True, None, False, 'string', object()])
    def test_not_a_dict(self, wrong_type):
        with pytest.raises(TypeError):
            sort_dict_by_values(wrong_type)

    @pytest.mark.parametrize('wrong_type', [[], set(), (), {}, 1, 1.0, True, False, 'string', object()])
    def test_key_is_not_a_function(self, wrong_type):
        with pytest.raises(TypeError):
            sort_dict_by_values({}, condition=wrong_type)

    @pytest.mark.parametrize('wrong_type', [[], set(), (), {}, 1, 1.0, None, 'string', object()])
    def test_not_a_bool(self, wrong_type):
        with pytest.raises(TypeError):
            sort_dict_by_values({}, reverse=wrong_type)

    def test_empty(self):
        assert sort_dict_by_values({}) == {}

    def test_simple(self):
        actual = sort_dict_by_values({'b': 2, 'c': 1, 'a': 3})
        expected = {'c': 1, 'b': 2, 'a': 3}
        assert actual == expected
        assert dicts_share_value_order(actual, expected)

    def test_with_key(self):
        actual = sort_dict_by_values({'qwerty': 1, 'x': 3, 'asd': -2}, condition=abs)
        expected = {'qwerty': 1, 'asd': -2, 'x': 3}
        assert actual == expected
        assert dicts_share_value_order(actual, expected)

    def test_reversed(self):
        actual = sort_dict_by_values({'b': 2, 'c': 1, 'a': 3}, reverse=True)
        expected = {'a': 3, 'b': 2, 'c': 1}
        assert actual == expected
        assert dicts_share_value_order(actual, expected)

    def test_with_key_and_reversed(self):
        actual = sort_dict_by_values({'qwerty': 1, 'x': 3, 'asd': -2}, condition=abs, reverse=True)
        expected = {'x': 3, 'asd': -2, 'qwerty': 1}
        assert actual == expected
        assert dicts_share_value_order(actual, expected)


# noinspection PyTypeChecker
class Test_sort_dict_by_keys_recursive:
    @pytest.mark.parametrize('wrong_type', [1, 1.0, True, None, False, 'string', object()])
    def test_not_an_iterable(self, wrong_type):
        with pytest.raises(TypeError):
            sort_dict_by_keys_recursive(wrong_type)

    @pytest.mark.parametrize('wrong_type', [[], set(), (), {}, 1, 1.0, True, False, 'string', object()])
    def test_key_is_not_a_function(self, wrong_type):
        with pytest.raises(TypeError):
            sort_dict_by_keys_recursive({}, condition=wrong_type)

    @pytest.mark.parametrize('wrong_type', [[], set(), (), {}, 1, 1.0, None, 'string', object()])
    def test_not_a_bool(self, wrong_type):
        with pytest.raises(TypeError):
            sort_dict_by_keys_recursive({}, reverse=wrong_type)

    def test_empty(self):
        assert sort_dict_by_keys_recursive({}) == {}

    @pytest.mark.parametrize('data, expected', [
        [{'x': 1, 'a': 2}, {'a': 2, 'x': 1}],
        [{'a': 2, 'x': 1}, {'a': 2, 'x': 1}],
    ])
    def test_simple(self, data, expected):
        actual = sort_dict_by_keys_recursive(data)
        assert actual == expected
        assert dicts_share_key_order(actual, expected, recursive=True)

    @pytest.mark.parametrize('data, expected', [
        [
            [{'x': 1, 'a': 2}, map(lambda x: x, (2, 1, {'x': 1, 'a': {'z': 1, 'y': 2}}))],
            [{'a': 2, 'x': 1}, (2, 1, {'a': {'y': 2, 'z': 1}, 'x': 1})],
        ],
        [
            [{'a': 2, 'x': 1}, map(lambda x: x, (2, 1, {'a': {'y': 2, 'z': 1}, 'x': 1}))],
            [{'a': 2, 'x': 1}, (2, 1, {'a': {'y': 2, 'z': 1}, 'x': 1})],
        ],
    ])
    def test_recursive(self, data, expected):
        actual = sort_dict_by_keys_recursive(data)
        actual[1] = tuple(actual[1])
        assert actual == expected
        assert dicts_share_key_order(actual, expected, recursive=True)

    @pytest.mark.parametrize('data, expected', [
        [
            [{'x': 1, 'a': 2}, map(lambda x: x, (2, 1, {'x': 1, 'a': {'z': 1, 'y': 2}}))],
            [{'x': 1, 'a': 2}, (2, 1, {'x': 1, 'a': {'y': 2, 'z': 1}})],
        ],
        [
            [{'a': 2, 'x': 1}, map(lambda x: x, (2, 1, {'a': {'y': 2, 'z': 1}, 'x': 1}))],
            [{'x': 1, 'a': 2}, (2, 1, {'x': 1, 'a': {'y': 2, 'z': 1}})],
        ],
    ])
    def test_recursive_with_key(self, data, expected):
        def condition(k):
            if k in ['x', 'y']:
                return 0
            else:
                return ord(k)

        actual = sort_dict_by_keys_recursive(data, condition=condition)
        actual[1] = tuple(actual[1])
        assert actual == expected
        assert dicts_share_key_order(actual, expected, recursive=True)

    @pytest.mark.parametrize('data, expected', [
        [
            [{'x': 1, 'a': 2}, map(lambda x: x, (2, 1, {'x': 1, 'a': {'z': 1, 'y': 2}}))],
            [{'x': 1, 'a': 2}, (2, 1, {'x': 1, 'a': {'z': 1, 'y': 2}})],
        ],
        [
            [{'a': 2, 'x': 1}, map(lambda x: x, (2, 1, {'a': {'y': 2, 'z': 1}, 'x': 1}))],
            [{'x': 1, 'a': 2}, (2, 1, {'x': 1, 'a': {'z': 1, 'y': 2}})],
        ],
    ])
    def test_recursive_reversed(self, data, expected):
        actual = sort_dict_by_keys_recursive(data, reverse=True)
        actual[1] = tuple(actual[1])
        assert actual == expected
        assert dicts_share_key_order(actual, expected, recursive=True)


# noinspection PyTypeChecker
class Test_sort_dict_by_values_recursive:
    @pytest.mark.parametrize('wrong_type', [1, 1.0, True, None, False, 'string', object()])
    def test_not_an_iterable(self, wrong_type):
        with pytest.raises(TypeError):
            sort_dict_by_values_recursive(wrong_type)

    @pytest.mark.parametrize('wrong_type', [[], set(), (), {}, 1, 1.0, True, False, 'string', object()])
    def test_key_is_not_a_function(self, wrong_type):
        with pytest.raises(TypeError):
            sort_dict_by_values_recursive({}, condition=wrong_type)

    @pytest.mark.parametrize('wrong_type', [[], set(), (), {}, 1, 1.0, None, 'string', object()])
    def test_not_a_bool(self, wrong_type):
        with pytest.raises(TypeError):
            sort_dict_by_values_recursive({}, reverse=wrong_type)

    def test_empty(self):
        assert sort_dict_by_values_recursive({}) == {}

    def test_simple(self):
        actual = sort_dict_by_values_recursive({'x': 2, 'a': 1})
        expected = {'a': 1, 'x': 2}
        assert actual == expected
        assert dicts_share_value_order(actual, expected, recursive=True)

    @pytest.mark.parametrize('data, expected', [
        [
            [{'a': 2, 'b': 1}, map(lambda x: x, (2, 1, {'a': 2, 'b': 1}))],
            [{'b': 1, 'a': 2}, (2, 1, {'b': 1, 'a': 2})],
        ],
        [
            [{'b': 1, 'a': 2}, map(lambda x: x, (2, 1, {'b': 1, 'a': 2}))],
            [{'b': 1, 'a': 2}, (2, 1, {'b': 1, 'a': 2})],
        ],
    ])
    def test_recursive(self, data, expected):
        actual = sort_dict_by_values_recursive(data)
        actual[1] = tuple(actual[1])
        assert actual == expected
        assert dicts_share_value_order(actual, expected, recursive=True)


class Test_sort_recursive:
    @pytest.mark.parametrize('wrong_type', [1, 1.0, True, None, False, 'string', object()])
    def test_data_should_be_an_iterable(self, wrong_type):
        with pytest.raises(TypeError):
            sort_recursive(wrong_type)

    @pytest.mark.parametrize('wrong_type', [
        iter([]), map(lambda x: x, []), range(1), object(), [], {}, (), set(), 1, True, False, None, ''
    ])
    def test_key_should_be_callable(self, wrong_type):
        with pytest.raises(TypeError):
            sort_recursive([], condition=wrong_type)

    @pytest.mark.parametrize('wrong_type', [[], set(), (), {}, 1, 1.0, None, 'string', object()])
    def test_reverse_should_be_a_bool(self, wrong_type):
        with pytest.raises(TypeError):
            sort_recursive({}, reverse=wrong_type)

    @pytest.mark.parametrize('wrong_value', [[], set(), (), {}, 1, 1.0, 'string', object(), True, False])
    def test_sort_dicts_by_be_keys_values_or_None(self, wrong_value):
        with pytest.raises(ValueError):
            sort_recursive({}, sort_dicts_by=wrong_value)

    @pytest.mark.parametrize('wrong_type', [[], set(), (), {}, 1, 1.0, None, 'string', object()])
    def test_sort_iters_should_be_bool(self, wrong_type):
        with pytest.raises(TypeError):
            sort_recursive({}, sort_iters=wrong_type)

    @pytest.mark.parametrize('wrong_type', [[], set(), (), {}, 1, 1.0, None, 'string', object()])
    def test_sort_lists_should_be_bool(self, wrong_type):
        with pytest.raises(TypeError):
            sort_recursive({}, sort_lists=wrong_type)

    @pytest.mark.parametrize('wrong_type', [[], set(), (), {}, 1, 1.0, None, 'string', object()])
    def test_sort_sets_should_be_bool(self, wrong_type):
        with pytest.raises(TypeError):
            sort_recursive({}, sort_sets=wrong_type)

    @pytest.mark.parametrize('wrong_type', [[], set(), (), {}, 1, 1.0, None, 'string', object()])
    def test_sort_tuples_should_be_bool(self, wrong_type):
        with pytest.raises(TypeError):
            sort_recursive({}, sort_tuples=wrong_type)

    @pytest.mark.parametrize('data, expected', [
        ([], []),
        ((), ()),
        ({}, {}),
        (set(), []),
        (iter([]), []),
        (map(bool, []), []),
    ])
    def test_sort_empty(self, data, expected):
        assert sort_recursive(data) == expected

    @pytest.mark.parametrize('data', [
        [],
        (),
        {},
        set(),
        iter([]),
        map(bool, []),
    ])
    def test_returns_same_type_if_possible(self, data):
        if isinstance(data, (list, tuple, dict)):
            assert isinstance(sort_recursive(data), type(data))
        elif isinstance(data, set) or is_iterator(data):
            assert isinstance(sort_recursive(data), list)

    @pytest.mark.parametrize('data, expected', [
        ([3, 2, 1], [1, 2, 3]),
        ((3, 2, 1), (1, 2, 3)),
        ({3: 'x', 2: 'x', 1: 'x'}, {1: 'x', 2: 'x', 3: 'x'}),
        ({3, 2, 1}, [1, 2, 3]),
        (iter([3, 2, 1]), [1, 2, 3]),
        (map(bool, [2, 0, 1]), [False, True, True]),
    ])
    def test_sort_simple(self, data, expected):
        assert sort_recursive(data) == expected

    @pytest.mark.parametrize('kwargs, expected', [
        ({'data': [3, 2, 1], 'sort_lists': False}, [3, 2, 1]),
        ({'data': (3, 2, 1), 'sort_tuples': False}, (3, 2, 1)),
        ({'data': {3: 'x', 2: 'x', 1: 'x'}, 'sort_dicts_by': None}, {3: 'x', 2: 'x', 1: 'x'}),
        ({'data': {3, 2, 1}, 'sort_sets': False}, {3, 2, 1}),
        ({'data': iter([3, 2, 1]), 'sort_iters': False}, iter([3, 2, 1])),
        ({'data': map(bool, [2, 0, 1]), 'sort_iters': False}, map(bool, [2, 0, 1])),
    ])
    def test_type_remains_unchanged_when_not_sorting(self, kwargs, expected):
        actual = sort_recursive(**kwargs)

        if is_iterator(kwargs['data']):
            assert is_iterator(actual)
            assert list(actual) == list(expected)
        else:
            assert isinstance(actual, type(kwargs['data']))
            assert actual == expected

    @pytest.mark.parametrize('kwargs, expected', [
        ({'data': [(3,5), (4,2), (6, 1)], 'sort_lists': False}, [(3, 5), (2, 4), (1, 6)]),
        ({'data': ([3,5], [4,2], [6, 1]), 'sort_tuples': False}, ([3, 5], [2, 4], [1, 6])),
    ])
    def test_sort_only_inner(self, kwargs, expected):
        actual = sort_recursive(**kwargs)
        if is_iterator(expected):
            assert list(actual) == list(expected)
        else:
            assert actual == expected

    @pytest.mark.parametrize('kwargs, expected', [
        ({'data': [(3,5), (4,2), (6, 1)]}, [(1, 6), (2, 4), (3, 5)]),
        ({'data': ([3,5], [4,2], [6, 1])}, ([1, 6], [2, 4], [3, 5])),
        ({'data': {'a': (2, 3), 'b': (4, 1)}, 'sort_dicts_by': 'values'}, {'b': (1, 4), 'a': (2, 3)}),
    ])
    def test_sort_recursive(self, kwargs, expected):
        if is_iterator(expected):
            assert list(sort_recursive(**kwargs)) == list(expected)
        else:
            assert sort_recursive(**kwargs) == expected


class Test_swap_keys_and_values:
    @pytest.mark.parametrize('wrong_type', all_types_besides('dict'))
    def test_wrong_type(self, wrong_type):
        with pytest.raises(TypeError):
            swap_keys_and_values(wrong_type)

    @pytest.mark.parametrize('data, expected', [
        ({1: 2}, {2: 1}),
        ({1: 'a', 3: 'b'}, {'a': 1, 'b': 3}),
        ({1: 2, 3: 2}, {2: 3}),
        ({}, {})
    ])
    def test_simple(self, data, expected):
        assert swap_keys_and_values(data) == expected


class Test_xor:
    @pytest.mark.parametrize('wrong_type', [
        iter([]), map(lambda x: x, []), range(1), object(), [], {}, (), set(), 1, True, False, None, ''
    ])
    def test_key_should_be_callable(self, wrong_type):
        with pytest.raises(TypeError):
            xor(1, 2, condition=wrong_type)

    @pytest.mark.parametrize('values', [
        [],
        [1],
    ])
    def test_at_least_2_values(self, values):
        with pytest.raises(ValueError):
            xor(*values)

    @pytest.mark.parametrize('values', [
        [0, None, False],
        [1, 2],
    ])
    def test_none(self, values):
        assert xor(*values) is None

    @pytest.mark.parametrize('values, expected', [
        ([0, None, 2, False], 2),
        ([1, 0], 1),
        ([0, True], True),
    ])
    def test_ok(self, values, expected):
        assert is_equal(xor(*values),  expected)


class Test_xor_with_idx:
    @pytest.mark.parametrize('wrong_type', [
        iter([]), map(lambda x: x, []), range(1), object(), [], {}, (), set(), 1, True, False, None, ''
    ])
    def test_key_should_be_callable(self, wrong_type):
        with pytest.raises(TypeError):
            xor_with_idx(1, 2, condition=wrong_type)

    @pytest.mark.parametrize('values', [
        [],
        [1],
    ])
    def test_at_least_2_values(self, values):
        with pytest.raises(ValueError):
            xor_with_idx(*values)

    @pytest.mark.parametrize('values', [
        [0, None, False],
        [1, 2],
    ])
    def test_none(self, values):
        assert xor_with_idx(*values) is None

    @pytest.mark.parametrize('values, expected', [
        ([0, None, 5, False], (2, 5)),
        ([1, 0], (0, 1)),
        ([0, True], (1, True)),
    ])
    def test_ok(self, values, expected):
        actual = xor_with_idx(*values)
        assert is_equal(actual, expected)


class Test_detect_fronts:
    @pytest.mark.parametrize('wrong_type', [
        iter([]), map(lambda x: x, []), range(1), object(), [], {}, (), set(), None, ''
    ])
    def test_key_should_be_callable(self, wrong_type):
        with pytest.raises(TypeError):
            detect_fronts([0, 1, wrong_type, True, False])

    @pytest.mark.parametrize('values, expected', [
        ([], []),
        ([0, 0, 0], []),
        ([1, 1, 1], []),
    ])
    def test_none(self, values, expected):
        assert detect_fronts(values) == expected

    def test_from_string(self):
        expected = [
            (3, 1),
            (4, -1),
            (6, 1),
            (7, -1),
            (8, 1),
            (11, -1),
            (12, 1),
            (14, -1),
        ]
        assert detect_fronts('000100101110110') == expected

    def test_from_iter(self):
        expected = [
            (3, 1),
            (4, -1),
            (6, 1),
            (7, -1),
            (8, 1),
            (11, -1),
            (12, 1),
            (14, -1),
        ]
        assert detect_fronts(map(int, '000100101110110')) == expected

    def test_detect_positive_fronts(self):
        assert detect_fronts_positive([0, 1, 0, 1]) == [1, 3]
        assert detect_fronts_positive([1, 0]) == []

    def test_detect_negative_fronts(self):
        assert detect_fronts_negative([1, 0, 1, 0]) == [1, 3]
        assert detect_fronts_negative([0, 1]) == []



import pytest
from katalytic.data import all_types_besides
from katalytic.maths import clip, L1, L2, min_max


class Test_L1:
    @pytest.mark.parametrize('x', all_types_besides(['int', 'float']))
    def test_wrong_value(self, x):
        with pytest.raises(ValueError):
            L1(x, 1)

        with pytest.raises(ValueError):
            L1(1, x)

    @pytest.mark.parametrize('a, b', [
        ([], []),
        ([], [1]),
        ([1, 2], [1]),
        ([[1]], [1])
    ])
    def test_wrong_format(self, a, b):
        with pytest.raises(ValueError):
            L1(a, b)

    @pytest.mark.parametrize('a, b, expected', [
        (1, 1, 0),
        (1, 5, 4),
        (-5, 1, 6),
        ([1], [3], 2),
        ([3], [1], 2),
        ([1, 2], [3, 4], 4),
        ([1.1, 2], [3, 3.5], 3.4),
        ([1, 10, 1], (10, 10, 10), 18),
    ])
    def test_ok(self, a, b, expected):
        assert L1(a, b) == expected


class Test_L2:
    @pytest.mark.parametrize('x', all_types_besides(['int', 'float']))
    def test_wrong_value(self, x):
        with pytest.raises(ValueError):
            L2(x, 1)

        with pytest.raises(ValueError):
            L2(1, x)

    @pytest.mark.parametrize('a, b', [
        ([], []),
        ([], [1]),
        ([1, 2], [1]),
        ([[1]], [1]),
        (['1'], [1]),
        ([True], [1]),
    ])
    def test_wrong_value_2(self, a, b):
        with pytest.raises(ValueError):
            L2(a, b)

    @pytest.mark.parametrize('a, b, expected', [
        (1, 1, 0),
        (1, 5, 4),
        (-5, 1, 6),
        ([1], [3], 2),
        ([3], [1], 2),
        ([1, 2], [5, 6], 5.656854),
        ([1, 10, 1], (10, 10, 10), 12.727922),
    ])
    def test_ok(self, a, b, expected):
        assert round(L2(a, b), 6) == expected


class Test_min_max:
    @pytest.mark.parametrize('wrong_type', all_types_besides(['iterables']))
    def test_not_an_iterable(self, wrong_type):
        with pytest.raises(TypeError):
            min_max(wrong_type)

    @pytest.mark.parametrize('wrong_type', all_types_besides(['callables', 'none']))
    def test_not_a_callable(self, wrong_type):
        with pytest.raises(TypeError):
            min_max([], key=wrong_type)

    @pytest.mark.parametrize('data', [
        [],
        (),
        iter([]),
    ])
    def test_empty_without_default(self, data):
        with pytest.raises(ValueError):
            min_max(data)

    @pytest.mark.parametrize('data, default', [
        ([], 'default_1'),
        ((), 'default_2'),
        (iter([]), 'default_3'),
    ])
    def test_empty_with_default(self, data, default):
        assert min_max(data, default=default) == default


    @pytest.mark.parametrize('data, expected', [
        ({3,2,1}, (1, 3)),
        ([3, 2, -1, 100], (-1, 100)),
    ])
    def test_happy_path(self, data, expected):
        assert min_max(data) == expected


class Test_clip:
    @pytest.mark.parametrize('wrong_type', all_types_besides('numbers'))
    def test_wrong_type_x(self, wrong_type):
        with pytest.raises(TypeError):
            clip(wrong_type, 1, 2)

    @pytest.mark.parametrize('wrong_type', all_types_besides('numbers'))
    def test_wrong_type_min(self, wrong_type):
        with pytest.raises(TypeError):
            clip(1, wrong_type, 2)

    @pytest.mark.parametrize('wrong_type', all_types_besides('numbers'))
    def test_wrong_type_max(self, wrong_type):
        with pytest.raises(TypeError):
            clip(1, 2, wrong_type)

    @pytest.mark.parametrize('args, expected', [
        ((2, 1, 3), 2),
        ((2.5, 2, 3), 2.5),
    ])
    def test_within_bounds(self, args, expected):
        assert clip(*args) == expected

    @pytest.mark.parametrize('args, expected', [
        ((20, 1, 3), 3),
        ((0.5, 2, 3), 2),
    ])
    def test_outside_bounds(self, args, expected):
        assert clip(*args) == expected



import pytest

from katalytic.data.checks import (
    contains_all_of, contains_any_of, contains_none_of, is_any_of, is_dict_of_sequences, is_generator,
    is_generator_expression, is_generator_function, is_iterable, is_iterable_or_str, is_iterator, is_none_of, is_number,
    is_primitive, is_sequence, is_sequence_of_dicts, is_sequence_of_sequences, is_sequence_or_str,
    is_sequence_of_dicts_uniform, is_dict_of_sequences_uniform, is_sequence_of_sequences_uniform, dicts_share_key_order,
    dicts_share_value_order, is_singleton, is_equal
)

from katalytic.data import all_types, all_types_besides


class Test_contains_all_of:
    @pytest.mark.parametrize('wrong_type', [object(), 1, True, False, None, ''])
    def test_haystack_not_iterable(self, wrong_type):
        with pytest.raises(TypeError):
            contains_all_of(wrong_type, [])

    @pytest.mark.parametrize('wrong_type', [object(), 1, True, False, None, ''])
    def test_needles_not_iterable(self, wrong_type):
        with pytest.raises(TypeError):
            contains_all_of([], wrong_type)

    @pytest.mark.parametrize('haystack, needles', [
        (range(10), {1, 3, 5}),
        ({True, False, None}, [True, None]),
        ({True, False, None}, [True, False]),
        ({True, False, None}, [False, None]),
    ])
    def test_True(self, haystack, needles):
        assert contains_all_of(haystack, needles)

    @pytest.mark.parametrize('haystack, needles', [
        (range(10), {True, 1, 3, 5}),
        ([True], {(), 'hello', True}),
        (range(10), [True, False]),
        (range(10), [100, 666]),
        ({None}, (True, False)),
    ])
    def test_False(self, haystack, needles):
        assert not contains_all_of(haystack, needles)


class Test_contains_any_of:
    @pytest.mark.parametrize('wrong_type', [object(), 1, True, False, None, ''])
    def test_haystack_not_iterable(self, wrong_type):
        with pytest.raises(TypeError):
            contains_any_of(wrong_type, [])

    @pytest.mark.parametrize('wrong_type', [object(), 1, True, False, None, ''])
    def test_needles_not_iterable(self, wrong_type):
        with pytest.raises(TypeError):
            contains_any_of([], wrong_type)

    @pytest.mark.parametrize('haystack, needles', [
        (range(10), [True, False]),
        (range(10), [100, 666]),
        ({None}, (True, False)),
    ])
    def test_False(self, haystack, needles):
        assert not contains_any_of(haystack, needles)

    @pytest.mark.parametrize('haystack, needles', [
        (range(10), {True, False, 5}),
        ([True], {(), 'hello', True}),
    ])
    def test_True(self, haystack, needles):
        assert contains_any_of(haystack, needles)


class Test_contains_none_of:
    @pytest.mark.parametrize('wrong_type', [object(), 1, True, False, None, ''])
    def test_haystack_not_iterable(self, wrong_type):
        with pytest.raises(TypeError):
            contains_none_of(wrong_type, [])

    @pytest.mark.parametrize('wrong_type', [object(), 1, True, False, None, ''])
    def test_needles_not_iterable(self, wrong_type):
        with pytest.raises(TypeError):
            contains_none_of([], wrong_type)

    @pytest.mark.parametrize('haystack, needles', [
        (range(10), [True, False]),
        (range(10), [100, 666]),
        ({None}, (True, False)),
    ])
    def test_True(self, haystack, needles):
        assert contains_none_of(haystack, needles)

    @pytest.mark.parametrize('haystack, needles', [
        (range(10), {True, False, 5}),
        ([True], {(), 'hello', True}),
    ])
    def test_False(self, haystack, needles):
        assert not contains_none_of(haystack, needles)


# noinspection PyTypeChecker
class Test_dicts_share_key_order:
    @pytest.mark.parametrize('wrong_type', [[], set(), (), {}, 1, 1.0, None, 'string', object()])
    def test_not_a_bool(self, wrong_type):
        with pytest.raises(TypeError):
            dicts_share_key_order({}, {}, recursive=wrong_type)

    @pytest.mark.parametrize('wrong_type', [[], set(), (), 1, 1.0, None, 'string', object()])
    def test_not_a_dict(self, wrong_type):
        with pytest.raises(TypeError):
            dicts_share_key_order({}, wrong_type)

        with pytest.raises(TypeError):
            dicts_share_key_order(wrong_type, wrong_type)

        with pytest.raises(TypeError):
            dicts_share_key_order({'a': {}}, {'a': wrong_type}, recursive=True)

    @pytest.mark.parametrize('wrong_type', [[], (), 1, 1.0, None, 'string', object()])
    def test_no_dicts(self, wrong_type):
        with pytest.raises(TypeError):
            dicts_share_key_order(wrong_type, wrong_type, recursive=False)

    @pytest.mark.parametrize('wrong_type', [{}, 1, 1.0, None, 'string', object()])
    def test_not_a_sequence(self, wrong_type):
        with pytest.raises(TypeError):
            dicts_share_key_order([], wrong_type, recursive=True)

        with pytest.raises(TypeError):
            dicts_share_key_order({'a': []}, {'a': wrong_type}, recursive=True)

    def test_empty(self):
        assert dicts_share_key_order({}, {})

    @pytest.mark.parametrize('dict_1, dict_2', [
        [{'a': 1, 'b': 2}, {'a': 3, 'b': 4}],
        [{'a': {'b': 1, 'c': 2}}, {'a': {'b': 1, 'c': 2}}],
        [[{'a': 1, 'b': 2}], [{'a': 3, 'b': 4}]],
    ])
    def test_True(self, dict_1, dict_2):
        assert dicts_share_key_order(dict_1, dict_2, recursive=True)

    @pytest.mark.parametrize('dict_1, dict_2', [
        [{'a': 1, 'b': 2}, {'b': 1, 'a': 2}],
        [{'a': 1, 'b': 2}, {'a': 1, 'c': 2}],
        [{'a': {'b': 1, 'c': 2}}, {'a': {'c': 1, 'b': 2}}],
        [{'a': {'b': 1, 'c': 2}}, {'a': {'x': 1, 'y': 2}}],
        [{'a': {'b': 1, 'c': 2}}, {'z': {'x': 1, 'y': 2}}],
        [[{'a': 1, 'b': 2}], [{'b': 3, 'a': 4}]],
        [[{'a': 1, 'b': 2}], [{'x': 3, 'y': 4}]],
    ])
    def test_False(self, dict_1, dict_2):
        assert not dicts_share_key_order(dict_1, dict_2, recursive=True)


# noinspection PyTypeChecker
class Test_dicts_share_value_order:
    @pytest.mark.parametrize('wrong_type', [[], set(), (), {}, 1, 1.0, None, 'string', object()])
    def test_not_a_bool(self, wrong_type):
        with pytest.raises(TypeError):
            dicts_share_value_order({}, {}, recursive=wrong_type)

    @pytest.mark.parametrize('wrong_type', [[], set(), (), 1, 1.0, None, 'string', object()])
    def test_not_a_dict(self, wrong_type):
        with pytest.raises(TypeError):
            dicts_share_value_order({}, wrong_type)

        with pytest.raises(TypeError):
            dicts_share_value_order(wrong_type, wrong_type)

    @pytest.mark.parametrize('wrong_type', [[], (), 1, 1.0, None, 'string', object()])
    def test_no_dicts(self, wrong_type):
        with pytest.raises(TypeError):
            dicts_share_value_order(wrong_type, wrong_type, recursive=False)

    @pytest.mark.parametrize('wrong_type', [{}, 1, 1.0, None, 'string', object()])
    def test_not_a_sequence(self, wrong_type):
        with pytest.raises(TypeError):
            dicts_share_value_order([], wrong_type)

    def test_empty(self):
        assert dicts_share_value_order({}, {})

    @pytest.mark.parametrize('dict_1, dict_2', [
        [{'a': 1, 'b': 2}, {'c': 1, 'd': 2}],
        [{'a': {'b': 1, 'c': 2}}, {'b': {'b': 1, 'c': 2}}],
        [[{'a': 1, 'b': 2}], [{'c': 1, 'd': 2}]],
    ])
    def test_simple(self, dict_1, dict_2):
        assert dicts_share_value_order(dict_1, dict_2, recursive=True)
        assert dicts_share_value_order(dict_1, dict_2, recursive=True)
        assert dicts_share_value_order(dict_1, dict_2, recursive=True)

    @pytest.mark.parametrize('dict_1, dict_2', [
        [{'a': {'b': 1, 'c': 2}}, {'b': {'c': 1, 'b': 2}}],
        [{'a': {'b': 1, 'c': 2}}, {'b': {'b': 2, 'c': 1}}],
        [{'a': 1, 'b': 2}, {'c': 2, 'd': 1}],
        [[{'a': 2, 'b': 1}], [{'c': 1, 'd': 2}]],
    ])
    def test_recursive(self, dict_1, dict_2):
        assert not dicts_share_value_order(dict_1, dict_2, recursive=True)


class Test_is_any_of:
    @pytest.mark.parametrize('wrong_type', [object(), 1, True, False, None, ''])
    def test_not_iterable(self, wrong_type):
        with pytest.raises(TypeError):
            is_any_of([], haystack=wrong_type)

    @pytest.mark.parametrize('x, iterable', [
        (0, [1, 2, 3]),
        (4, [1, 2, 3]),
        ('a', {'b': 'a', 'c': 'd'}),
        ({}, [1, 2, 3]),
        (True, {1, None, False}),
    ])
    def test_False(self, x, iterable):
        assert not is_any_of(x, iterable)

    @pytest.mark.parametrize('x, iterable', [
        (2, [1, 2, 3]),
        ('a', {'a': 'b', 'c': 'd'}),
        ({}, [1, 2, {}, 3]),
        (True, {None, True, False}),
    ])
    def test_True(self, x, iterable):
        assert is_any_of(x, iterable)

    @pytest.mark.parametrize('x, iterable', [
        (1, [True, False]),
        (0, [True, False]),
        (True, [0, 1]),
        (False, [0, 1]),
    ])
    def test_bools_are_not_ints(self, x, iterable):
        assert not is_any_of(x, iterable)


class Test_is_dict_of_sequences:
    @pytest.mark.parametrize('correct_type', [{'a': []}, {'a': ()}])
    def test_True(self, correct_type):
        assert is_dict_of_sequences(correct_type)

    @pytest.mark.parametrize('wrong_type', all_types_besides('dict') + [{'a': {}}])
    def test_False(self, wrong_type):
        assert not is_dict_of_sequences(wrong_type)

    def test_empty_ok(self):
        assert is_dict_of_sequences({}, empty_ok=True)
        assert not is_dict_of_sequences({}, empty_ok=False)

    @pytest.mark.parametrize('mistake', all_types_besides('booleans'))
    def test_precondition_empty_ok(self, mistake):
        with pytest.raises(TypeError):
            is_dict_of_sequences([], empty_ok=mistake)


class Test_is_dict_of_sequences_uniform:
    @pytest.mark.parametrize('data', [
        {'a': (), 'b': ()},
        {'a': [1], 'b': [1]},
        {'a': (1, 2, 3), 'b': [10, 20, 30]},
    ])
    def test_True(self, data):
        assert is_dict_of_sequences_uniform(data)

    @pytest.mark.parametrize('data', [
        {'a': [1], 'b': []},
        {'a': [1], 'b': [1, 2]},
        {'a': [1], 'b': [1], 'c': [1, 2]},
        []
    ])
    def test_False(self, data):
        assert not is_dict_of_sequences_uniform(data)

    def test_empty_ok(self):
        assert is_dict_of_sequences_uniform({}, empty_ok=True)
        assert not is_dict_of_sequences_uniform({}, empty_ok=False)

    @pytest.mark.parametrize('mistake', all_types_besides('booleans'))
    def test_precondition_empty_ok(self, mistake):
        with pytest.raises(TypeError):
            is_dict_of_sequences_uniform([], empty_ok=mistake)


class Test_is_equal:
    @pytest.mark.parametrize('a, b', [
        (0, False),
        (0, None),
        (False, None),
        (1, True),
        (0, 1),
        (1, '1'),
    ])
    def test_False(self, a, b):
        assert not is_equal(a, b)

    @pytest.mark.parametrize('a, b', [
        (0, 0),
        (1, 1),
        (False, False),
        (True, True),
        (None, None),
    ])
    def test_True(self, a, b):
        assert is_equal(a, b)


class Test_is_generator:
    @pytest.mark.parametrize('correct_type', all_types('generators'))
    def test_True(self, correct_type):
        assert is_generator(correct_type)

    @pytest.mark.parametrize('wrong_type', all_types_besides('generators'))
    def test_False(self, wrong_type):
        assert not is_generator(wrong_type)


class Test_is_generator_function:
    @pytest.mark.parametrize('correct_type', all_types('generator_function'))
    def test_True(self, correct_type):
        assert is_generator_function(correct_type)

    @pytest.mark.parametrize('wrong_type', all_types_besides('generator_function'))
    def test_False(self, wrong_type):
        assert not is_generator_function(wrong_type)


class Test_is_generator_expression:
    @pytest.mark.parametrize('correct_type', all_types('generator_expression'))
    def test_True(self, correct_type):
        assert is_generator_expression(correct_type)

    @pytest.mark.parametrize('wrong_type', all_types_besides('generator_expression'))
    def test_False(self, wrong_type):
        assert not is_generator_expression(wrong_type)


class Test_is_iterable:
    @pytest.mark.parametrize('correct_type', all_types('iterables'))
    def test_True(self, correct_type):
        assert is_iterable(correct_type)

    @pytest.mark.parametrize('wrong_type', all_types_besides('iterables'))
    def test_False(self, wrong_type):
        assert not is_iterable(wrong_type)


class Test_is_iterable_or_str:
    @pytest.mark.parametrize('correct_type', all_types(['iterables', 'strings']))
    def test_True(self, correct_type):
        assert is_iterable_or_str(correct_type)

    @pytest.mark.parametrize('wrong_type', all_types_besides(['iterables', 'strings']))
    def test_False(self, wrong_type):
        assert not is_iterable_or_str(wrong_type)


class Test_is_iterator:
    @pytest.mark.parametrize('correct_type', all_types('iterators'))
    def test_True(self, correct_type):
        assert is_iterator(correct_type)

    @pytest.mark.parametrize('wrong_type', all_types_besides('iterators'))
    def test_False(self, wrong_type):
        assert not is_iterator(wrong_type)


class Test_is_none_of:
    @pytest.mark.parametrize('wrong_type', [object(), 1, True, False, None, ''])
    def test_not_iterable(self, wrong_type):
        with pytest.raises(TypeError):
            is_none_of([], haystack=wrong_type)

    @pytest.mark.parametrize('x, iterable', [
        (0, [1, 2, 3]),
        (4, [1, 2, 3]),
        ('a', {'b': 'a', 'c': 'd'}),
        ({}, [1, 2, 3]),
        (True, {1, None, False}),
    ])
    def test_True(self, x, iterable):
        assert is_none_of(x, iterable)

    @pytest.mark.parametrize('x, iterable', [
        (2, [1, 2, 3]),
        ('a', {'a': 'b', 'c': 'd'}),
        ({}, [1, 2, {}, 3]),
        (True, {None, True, False}),
    ])
    def test_False(self, x, iterable):
        assert not is_none_of(x, iterable)

    @pytest.mark.parametrize('x, iterable', [
        (1, [True, False]),
        (0, [True, False]),
        (True, [0, 1]),
        (False, [0, 1]),
    ])
    def test_bools_are_not_ints(self, x, iterable):
        assert is_none_of(x, iterable)


class Test_is_primitive:
    @pytest.mark.parametrize('correct_type', all_types('primitives'))
    def test_True(self, correct_type):
        assert is_primitive(correct_type)

    @pytest.mark.parametrize('wrong_type', all_types_besides('primitives'))
    def test_False(self, wrong_type):
        assert not is_primitive(wrong_type)


class Test_is_sequence:
    @pytest.mark.parametrize('correct_type', all_types('sequences'))
    def test_True(self, correct_type):
        assert is_sequence(correct_type)

    @pytest.mark.parametrize('wrong_type', all_types_besides('sequences'))
    def test_False(self, wrong_type):
        assert not is_sequence(wrong_type)


class Test_is_sequence_of_dicts:
    @pytest.mark.parametrize('mistake', all_types_besides('booleans'))
    def test_precondition_empty_ok(self, mistake):
        with pytest.raises(TypeError):
            is_sequence_of_dicts([], empty_ok=mistake)

    @pytest.mark.parametrize('correct_type', [[{}], ({}, )])
    def test_True(self, correct_type):
        assert is_sequence_of_dicts(correct_type)

    @pytest.mark.parametrize('wrong_type', all_types_besides('sequences') + [{'a': []}])
    def test_False(self, wrong_type):
        assert not is_sequence_of_dicts(wrong_type)

    def test_empty_ok(self):
        assert is_sequence_of_dicts([], empty_ok=True)
        assert not is_sequence_of_dicts([], empty_ok=False)


class Test_is_sequence_of_dicts_uniform:
    @pytest.mark.parametrize('mistake', all_types_besides('booleans'))
    def test_precondition_empty_ok(self, mistake):
        with pytest.raises(TypeError):
            is_sequence_of_dicts_uniform([], empty_ok=mistake)

    @pytest.mark.parametrize('data', [
        [{}, {}],
        [{'a': 1}, {'a': 2}],
        [{'b': 1, 'a': 1}, {'a': 2, 'b': 2}],
    ])
    def test_True(self, data):
        assert is_sequence_of_dicts_uniform(data)

    @pytest.mark.parametrize('data', [
        [{'a': 1}, {'b': 2}],
        [{'a': 1, 'b': 1}, {'b': 2}],
        {},
    ])
    def test_False(self, data):
        assert not is_sequence_of_dicts_uniform(data)

    def test_empty_ok(self):
        assert is_sequence_of_dicts_uniform([], empty_ok=True)
        assert not is_sequence_of_dicts_uniform([], empty_ok=False)


class Test_is_sequence_of_sequences:
    @pytest.mark.parametrize('mistake', all_types_besides('booleans'))
    def test_precondition_empty_ok(self, mistake):
        with pytest.raises(TypeError):
            is_sequence_of_sequences([], empty_ok=mistake)

    @pytest.mark.parametrize('correct_type', [[[]], [()], ([], ), ((), )])
    def test_True(self, correct_type):
        assert is_sequence_of_sequences(correct_type)

    @pytest.mark.parametrize('wrong_type', all_types_besides('sequences') + [{'a': []}])
    def test_False(self, wrong_type):
        assert not is_sequence_of_sequences(wrong_type)

    def test_empty_ok(self):
        assert is_sequence_of_sequences([], empty_ok=True)
        assert not is_sequence_of_sequences([], empty_ok=False)


class Test_is_sequence_of_sequences_uniform:
    @pytest.mark.parametrize('mistake', all_types_besides('booleans'))
    def test_precondition_empty_ok(self, mistake):
        with pytest.raises(TypeError):
            is_sequence_of_sequences_uniform([], empty_ok=mistake)

    @pytest.mark.parametrize('data', [
        [[], []],
        [
            ['a', 'b', 'c'],
            ['x', 'y', 'z'],
            [1, 2, 3],
            [True, False, None]
        ],
    ])
    def test_True(self, data):
        assert is_sequence_of_sequences_uniform(data)

    @pytest.mark.parametrize('data', [
        [[], [1]],
        [['a', 'b', 'c'], ['x', 'y']],
        {},
    ])
    def test_False(self, data):
        assert not is_sequence_of_sequences_uniform(data)

    def test_empty_ok(self):
        assert is_sequence_of_sequences_uniform([], empty_ok=True)
        assert not is_sequence_of_sequences_uniform([], empty_ok=False)


class Test_is_sequence_or_str:
    @pytest.mark.parametrize('correct_type', all_types(['sequences', 'strings']))
    def test_True(self, correct_type):
        assert is_sequence_or_str(correct_type)

    @pytest.mark.parametrize('wrong_type', all_types_besides(['sequences', 'strings']))
    def test_False(self, wrong_type):
        assert not is_sequence_or_str(wrong_type)


class Test_is_singleton:
    @pytest.mark.parametrize('correct_type', all_types('singletons'))
    def test_True(self, correct_type):
        assert is_singleton(correct_type)

    @pytest.mark.parametrize('wrong_type', all_types_besides('singletons'))
    def test_False(self, wrong_type):
        assert not is_singleton(wrong_type)


class Test_is_number:
    @pytest.mark.parametrize('correct_type', all_types('numbers'))
    def test_True(self, correct_type):
        assert is_number(correct_type)

    @pytest.mark.parametrize('wrong_type', all_types_besides('numbers'))
    def test_False(self, wrong_type):
        assert not is_number(wrong_type)



import re
import sys
import traceback
import warnings

from glob import iglob
from pathlib import Path
from importlib import import_module


def __get_version_from_metadata(path):
    with open(path, 'r') as f:
        for line in f:
            if line.startswith('Version:'):
                return line.split(':')[-1].strip()


def __find_paths(pkg):
    pkg = pkg.replace('-', '_').replace('.', '_')
    for p in sys.path:
        if not Path(p).is_dir():
            continue

        for p2 in iglob(f'{p}/**.dist-info', recursive=True):
            if re.search(f'{pkg}-[^/]*[.]dist-info', p2):
                yield p2
                if Path(f'{p2}/METADATA').is_file():
                    yield p2

        for p2 in iglob(f'{p}/**.egg-info', recursive=True):
            if re.search(f'{pkg}-[^/]*[.]egg-info', p2):
                yield p2
                if Path(f'{p2}/PKG-INFO').is_file():
                    yield f'{p2}/PKG-INFO'


class KatalyticInterrupt(Exception):
    """
    This exception is used in testing how code behaves when it gets interrupted.
    I can catch all exceptions and re-raise if it's not this one
    """


def get_version(dunder_name):  # pragma: no cover -- I can't test all branches at the same time
    """
    Get the version information for a package.

    Args:
        dunder_name (str):
            The dunder name of the package (i.e., __name__).

    Returns:
        tuple:
            A tuple containing the version string and the version information as a tuple of integers.
            If the version information cannot be determined, None is returned.

    """
    if sys.version_info >= (3, 8):
        from importlib import metadata
        dunder_name = dunder_name.replace('.', '-')
        dunder_name = dunder_name.replace('-__init__', '')
        version = metadata.version(dunder_name)
        version_info = tuple(map(int, version.split('.')))
        return version, version_info

    version = None
    for p in __find_paths(dunder_name):
        if p.endswith('.dist-info'):
            version = re.search(r'\w+-(\d+\.\d+\.\d+)', p)
            if version:
                version = version.group(1)
                break

        if p.endswith('.dist-info/METADATA'):
            version = __get_version_from_metadata(p)
            if version:
                break

        if p.endswith('.egg-info/PKG-INFO'):
            version = __get_version_from_metadata(p)
            if version:
                break

    if version:
        version_info = tuple(map(int, version.split('.')))
    else:
        version_info = None

    # If nothing worked, I would rather return None than raise an exception
    # This could happen if the package is installed on python < 3.8
    return version, version_info


# __version__, __version_info__ = get_version(__name__)


def _get_stacktrace(e):
    return ''.join(traceback.format_exception(None, e, e.__traceback__))


def _check(name, group):
    if group.endswith('*'):
        prefix = group.split('*')[0]
        return name.startswith(prefix)
    else:
        return name == group


def get_modules():
    """
    Get a list of importable modules within the package.

    Returns:
        list:
            A list of importable modules within the package, sorted by module name.

    """
    modules = []
    for item in Path(__file__).parent.iterdir():
        if item.stem == '__pycache__':
            continue

        module_name = f'{__package__}.{item.stem}'
        try:
            module = import_module(module_name)
        except Exception as e:  # pragma: no cover
            warnings.warn(f'Couldn\'t import {module_name!r}\n' + _get_stacktrace(e))
            continue

        modules.append(module)

    # find editable installs
    for item in Path(__file__).parent.parent.iterdir():  # pragma: no cover
        # no coverage: because I would have to create an editable install,
        # just to test this functionality
        if not re.search(__package__ + r'\..*\.pth', item.name):
            continue

        try:
            module = import_module(item.stem)
        except Exception as e:  # pragma: no cover
            warnings.warn(f'Couldn\'t import {item.stem!r}\n' + _get_stacktrace(e))
            continue

        modules.append(module)

    return sorted(modules, key=lambda m: m.__name__)


def find_functions_marked_with(group):
    """
    Get a list of functions within the package that belong to a specific group.

    Args:
        group (str):
            The name of the group.

    Returns:
        list:
            A list of tuples containing the function name, the function object, and the groups it belongs to.
            The list is sorted based on the function names.

    """
    functions = []
    for module in get_modules():
        for func_name in dir(module):
            f = getattr(module, func_name)
            groups = getattr(f, '__katalytic_marks__', [])
            groups = [g for g in groups if _check(g, group)]
            if groups:
                functions.append((func_name, f, groups))

    return sorted(functions)


def mark(label):
    """
    Decorator to mark a function with one or more labels.

    Args:
        label (str):
            The label to mark the function with.

    Returns:
        callable:
            The decorator to mark a function.

    Raises:
        TypeError:
            If the provided label is not a string.
        ValueError:
            If the provided label contains newline characters, tab characters, or consists only of whitespace characters.

    """
    if not isinstance(label, str):
        raise TypeError(f'Only strings are allowed. Got {label!r}')

    if '\n' in label or '\t' in label or re.search(r'^\s*$', label):
        raise ValueError(f'Choose a meaningful label. Got {label!r}')

    def decorator(func):
        func.__katalytic_marks__ = getattr(func, '__katalytic_marks__', ())
        # prepend to maintain the intuitive order (top to bottom)
        func.__katalytic_marks__ = (label, *func.__katalytic_marks__)
        return func

    return decorator


@mark('__test_1')
@mark('__test_2')
@mark('__test_300')
def __test(): pass


@mark('__test_3::a')
@mark('__test_3::b')
@mark('__test_2')
def __test_2(): pass


_UNDEFINED = object()



# noinspection PyUnresolvedReferences
"""
This module contains various utility functions for extracting and transforming data.

Examples:
    >>> first([1, 6, 2], condition=lambda x: x % 2 == 0)
    6

    >>> detect_fronts_positive('00101101')
    [2, 4, 7]

    >>> swap_keys_and_values({'a': 1, 'b': 2, 'c': 3})
    {1: 'a', 2: 'b', 3: 'c'}

    >>> as_dict_of_lists([['a', 'b'], [1, 2], [3, 4]])
    {'a': [1, 3], 'b': [2, 4]}

"""

import copy
from decimal import Decimal
from fractions import Fraction
from pathlib import Path

from katalytic.data.checks import (
    is_any_of, is_iterable, is_iterator, is_none_of, is_primitive, is_dict_of_sequences_uniform, is_sequence,
    is_sequence_of_sequences_uniform, is_sequence_of_dicts_uniform
)
# noinspection PyProtectedMember
from katalytic._pkg import _UNDEFINED


def _function(): pass
def _generator(): yield 1


class _C:
    def __call__(self, *args, **kwargs):
        pass


_map = map(int, [])
_C_obj = _C()
_obj = object()
_lambda = lambda x: x
_generator_expr = (x for x in [])
_sequences = [[], (), range(0)]
_dict_views = [{}.keys(), {}.values(), {}.items()]
_iterators = [_generator, _generator_expr, iter([]), _map, enumerate([]), zip([], [])]
_iterables = [*_dict_views, *_iterators, *_sequences, set(), {}]
_collections = [(), set(), frozenset([]), {}, []]
_booleans = [True, False]
_singletons = [None, True, False]
_primitives = [*_singletons, 0, 0.0, '', b'', bytearray(b'')]
_callables = [_generator, _function, _lambda, _C_obj, _C]
_numbers = [0, 0.0, 0j, Decimal('0'), Fraction(0, 1)]
_objects = [_obj, _C_obj]
_generators = [_generator, _generator_expr]
_functions = [_generator, _function, _lambda]
_strings = ['', b'', bytearray(b'')]

_types = {
    'booleans': _booleans,
    'bytearray': bytearray(b''),
    'bytes': b'',
    'callables': _callables,
    'callable_obj': _C_obj,
    'class': _C,
    'collections': _collections,
    'complex': 0 + 0j,
    'decimal': Decimal('0'),
    'dict': {},
    'dict_views': _dict_views,
    'float': 0.0,
    'fraction': Fraction(0, 1),
    'frozenset': frozenset([]),
    'functions': _functions,
    'generator_expression': _generator_expr,
    'generator_function': _generator,
    'generators': _generators,
    'int': 0,
    'iterables': _iterables,
    'iterators': _iterators,
    'list': [],
    'map': _map,
    'none': None,
    'numbers': _numbers,
    'objects': _objects,
    'path': Path(''),
    'primitives': _primitives,
    'sequences': _sequences,
    'set': set(),
    'singletons': _singletons,
    'str': '',
    'strings': _strings,
    'tuple': (),
}


def all_types(whitelist=None):
    """
    Get all supported types or a subset of types based on a whitelist.

    Args:
        whitelist (str or Iterable[str], optional):
            A whitelist of type names to include. If None, all supported types are returned.
            If a string, it is treated as a single type name.
            If an iterable of strings, it is treated as a collection of type names.
            Defaults to None.

    Returns:
        list: A list of values for each type in the whitelist.

    Raises:
        TypeError: If whitelist is not None, a str, or an iterable.
        ValueError: If unexpected type names are present in the whitelist.

    Examples:
        >>> all_types('numbers')
        [0, 0.0, 0j, Decimal('0'), Fraction(0, 1)]
        >>> all_types('singletons')
        [None, True, False]

    """
    if whitelist is None:
        return _flatten(_types.values())
    elif isinstance(whitelist, str):
        whitelist = [whitelist]
    elif not is_iterable(whitelist):
        raise TypeError(f'<whitelist> must be iterable. Got {type(whitelist).__name__}')

    unexpected = set(whitelist) - set(_types.keys())
    if unexpected:
        raise ValueError(f'Unexpected types in <whitelist>: {unexpected}')

    return _flatten(_types[t] for t in whitelist)


def all_types_besides(blacklist):
    """
    Get all supported types except those specified in the blacklist.

    Args:
        blacklist (str or Iterable[str]):
            A list of type names to exclude. If a string, it is treated as a single type name.
            If an iterable of strings, it is treated as a collection of type names.

    Returns:
        list: A list of values for each type not in the blacklist.

    Raises:
        TypeError: If blacklist is not a str or iterable.
        ValueError: If unexpected type names are present in the blacklist.

    Examples:
        >>> all_types_besides(['iterables', 'generators', 'functions', 'objects', 'path', 'numbers', 'class'])
        [True, bytearray(b''), b'', None, '']

    """
    if isinstance(blacklist, str):
        blacklist = [blacklist]
    elif not is_iterable(blacklist):
        raise TypeError(f'<blacklist> must be iterable. Got {type(blacklist).__name__}')

    blacklist = set(blacklist)
    unexpected = blacklist - set(_types.keys())
    if unexpected:
        raise ValueError(f'Unexpected types in <blacklist>: {unexpected}')

    to_remove = _flatten(_types[t] for t in blacklist)
    all_types = _flatten(_types.values())
    kept = []
    for t in all_types:
        if t in to_remove:
            continue

        # remove duplicates too
        # I have to do it this way because python considers
        # 0 == 0.0 == 0j == Decimal('0') == Fraction(0, 1)
        if (t, type(t)) in [(x, type(x)) for x in kept]:
            continue

        kept.append(t)

    return kept


def _flatten(iterable):
    """
    Flatten iterable with some special rules to make all_types() and
    all_types_besides() work correctly. Use flatten() for everything else

    Args:
        iterable (Iterable): The iterable to flatten.

    Returns:
        list: An iterable flattened by one level.

    Raises:
        TypeError: If the input is not an iterable.

    """
    if not is_iterable(iterable):
        raise TypeError(f'<iterable> expects an iterable. Got {type(iterable).__name__}')

    flat = []
    for x in iterable:
        if isinstance(x, (dict, set, list, tuple)) and len(x):
            flat.extend(x)
        else:
            flat.append(x)

    return flat


def as_dict_of_lists(data, *, empty_ok=True):
    """
    Convert data into a dictionary of lists format.

    This format is useful when you need to perform operations on each column
    of a table. It provides a compact representation of the data, but may be less intuitive.

    Args:
        data (Union[Sequence[Dict], Sequence[Sequence], Dict]): The input data to convert.
        empty_ok (bool): Whether to allow an empty collection. Defaults to True.

    Returns:
        Dict: The data in dictionary of lists format.

    Raises:
        TypeError: If the input data has an unexpected format.

    Examples:
        >>> as_dict_of_lists({'b': [3, 4], 'a': [1, 2]})
        {'b': [3, 4], 'a': [1, 2]}

        >>> as_dict_of_lists(({'a': 1, 'b': 2}, {'b': 4, 'a': 3}))
        {'a': [1, 3], 'b': [2, 4]}

        >>> as_dict_of_lists([['b', 'a'], [1, 2], [3, 4]])
        {'b': [1, 3], 'a': [2, 4]}

    """
    if not isinstance(empty_ok, bool):
        raise TypeError(f'<empty_ok> expects False or True. Got {type(empty_ok)}')

    if (isinstance(data, dict) or is_sequence(data)) and len(data) == 0:
        if empty_ok:
            return []
        else:
            raise ValueError(f'Empty collection not allowed when <empty_ok> is False')
    elif is_sequence_of_dicts_uniform(data):
        return {k: [d[k] for d in data] for k in data[0]}
    elif is_sequence_of_sequences_uniform(data):
        return {k: [v[i] for v in data[1:]] for i, k in enumerate(data[0])}
    elif is_dict_of_sequences_uniform(data):
        return {k: list(v) for k, v in data.items()}
    else:
        raise TypeError(f'Unexpected format for <data>. Got {type(data).__name__}: {data!r}')


def as_list_of_dicts(data, *, empty_ok=True):
    """
    Convert data into a list of dictionaries format.

    This format is useful when you need to perform operations on each row of a table.

    Args:
        data (Union[Sequence[Dict], Sequence[Sequence], Dict]): The input data to convert.
        empty_ok (bool): Whether to allow an empty collection. Defaults to True.

    Returns:
        List[Dict]: The data in list of dictionaries format.

    Raises:
        TypeError: If the input data has an unexpected format.

    Examples:
        >>> as_list_of_dicts({'a': [1, 2], 'b': [3, 4]})
        [{'a': 1, 'b': 3}, {'a': 2, 'b': 4}]

        >>> as_list_of_dicts([['b', 'a'], [1, 2], [3, 4]])
        [{'b': 1, 'a': 2}, {'b': 3, 'a': 4}]

        >>> as_list_of_dicts([{'b': 2, 'a': 1}, {'a': 3, 'b': 4}])
        [{'b': 2, 'a': 1}, {'b': 4, 'a': 3}]

    """
    if not isinstance(empty_ok, bool):
        raise TypeError(f'<empty_ok> expects False or True. Got {type(empty_ok)}')

    if (isinstance(data, dict) or is_sequence(data)) and len(data) == 0:
        if empty_ok:
            return []
        else:
            raise ValueError(f'Empty collection not allowed when <empty_ok> is False')
    elif is_sequence_of_dicts_uniform(data):
        header = list(data[0].keys())
        return [{k: d[k] for k in header} for d in data]
    elif is_sequence_of_sequences_uniform(data):
        return [dict(zip(data[0], row)) for row in data[1:]]
    elif is_dict_of_sequences_uniform(data):
        first_seq = list(data.values())[0]
        seq_len = len(first_seq)
        return [{k: v[i] for k, v in data.items()} for i in range(seq_len)]
    else:
        raise TypeError(f'Unexpected format for <data>. Got {type(data).__name__}: {data!r}')


def as_list_of_lists(data, *, empty_ok=True):
    """
    Convert data into a list of lists format.

    This format, along with the dict_of_sequences format, is the most compact for storing tabular data.
    This format is more intuitive than the dict_of_sequences format.

    Args:
        data (Union[Sequence[Dict], Sequence[Sequence], Dict]): The input data to convert.
        empty_ok (bool): Whether to allow an empty collection. Defaults to True.

    Returns:
        List[List]: The data in list of lists format.

    Raises:
        TypeError: If the input data has an unexpected format.

    Examples:
        >>> as_list_of_lists({'a': [1, 2], 'b': [3, 4]})
        [['a', 'b'], [1, 3], [2, 4]]

        >>> as_list_of_lists(({'a': 1, 'b': 2}, {'a': 3, 'b': 4}))
        [['a', 'b'], [1, 2], [3, 4]]

        >>> as_list_of_lists([['b', 'a'], [1, 2], [3, 4]])
        [['b', 'a'], [1, 2], [3, 4]]

    """
    if not isinstance(empty_ok, bool):
        raise TypeError(f'<empty_ok> expects False or True. Got {type(empty_ok)}')

    if (isinstance(data, dict) or is_sequence(data)) and len(data) == 0:
        if empty_ok:
            return []
        else:
            raise ValueError(f'Empty collection not allowed when <empty_ok> is False')
    elif is_sequence_of_dicts_uniform(data):
        header = [list(data[0].keys())]
        rows = [[d[k] for k in header[0]] for d in data]
        return header + rows
    elif is_sequence_of_sequences_uniform(data):
        return list(map(list, data))
    elif is_dict_of_sequences_uniform(data):
        header = [list(data.keys())]
        n = len(list(data.values())[0])
        rows = [[v[i] for v in data.values()] for i in range(n)]
        return header + rows
    else:
        raise TypeError(f'Unexpected format for <data>. Got {type(data).__name__}: {data!r}')


def detect_fronts(bits):
    """
    Detects the fronts in a sequence of bits.

    A front is a change from 0 to 1 (positive) or from 1 to 0 (negative).
    It works even if the bits are booleans instead of 0/1.

    Returns a list of tuples
    Args:
        bits (Iterable or str): A string or a list representing binary values.

    Returns:
        List[Tuple[int, int]]: The list of tuples representing the fronts
            formatted as (index, change) where index is the index at which
            the bit flip takes place and change is 1 for positive fronts
            and -1 for negative fronts.

    Raises:
        TypeError: If the bits are not of the expected types.

    Example:
        >>> bits = "00101101"
        >>> detect_fronts(bits)
        [(2, 1), (3, -1), (4, 1), (6, -1), (7, 1)]

        >>> bits = [0, 0, 1, 0, 1, 1, 0, 1]
        >>> detect_fronts(bits)
        [(2, 1), (3, -1), (4, 1), (6, -1), (7, 1)]

        >>> bits = [False, False, True, False, True, True, False, True]
        >>> detect_fronts(bits)
        [(2, 1), (3, -1), (4, 1), (6, -1), (7, 1)]

    """
    if is_iterator(bits):
        bits = list(bits)
    elif isinstance(bits, str):
        bits = list(map(int, bits))

    if not all(is_any_of(b, (0, 1, True, False)) for b in bits):
        raise TypeError(f'Only 0/1 or True/False are allowed. Got {bits!r}')

    fronts = []
    for i, (a, b) in enumerate(zip(bits, bits[1:])):
        if (a, b) == (0, 1):  # it works for (False, True) as well
            fronts.append((i+1, 1))
        elif (a, b) == (1, 0):  # it works for (True, False) as well
            fronts.append((i+1, -1))

    return fronts


def detect_fronts_positive(bits):
    """Detects the positive fronts (a 0 to 1 transition).

    Args:
        bits (str or list): A string or a list representing binary values.

    Returns:
        list: A list of indices at which the bit flip from 0 to 1 occurs.

    Example:
        >>> detect_fronts_positive("00101101")
        [2, 4, 7]

        >>> detect_fronts_positive([0, 0, 1, 0, 1, 1, 0, 1])
        [2, 4, 7]

        >>> bits = [False, False, True, False, True, True, False, True]
        >>> detect_fronts_positive(bits)
        [2, 4, 7]

    """
    return [i for i, change in detect_fronts(bits) if change == 1]


def detect_fronts_negative(bits):
    """
    Detects the negative fronts (a 1 to 0 transition).

    Args:
        bits (str or list): A string or a list representing binary values.

    Returns:
        list: A list of indices at which the bit flip from 1 to 0 occurs.

    Example:
        >>> bits = "00101101"
        >>> detect_fronts_negative(bits)
        [3, 6]

        >>> bits = [0, 0, 1, 0, 1, 1, 0, 1]
        >>> detect_fronts_negative(bits)
        [3, 6]

        >>> bits = [False, False, True, False, True, True, False, True]
        >>> detect_fronts_negative(bits)
        [3, 6]

    """
    return [i for i, change in detect_fronts(bits) if change == -1]


def first(data, *, condition=lambda _: True):
    """
    Returns the first element from a sequence.

    Args:
        data (sequence): The sequence from which to retrieve the first element.
        condition (function, optional): A function to specify the condition for comparison. Defaults to lambda _: True.

    Returns:
        The first element from the sequence.

    Raises:
        TypeError: If data is a set. Use `one(data)` instead for sets.

    Example:
        >>> data = [1, 2, 3, 4, 5]
        >>> first(data)
        1

        >>> data = [1, 2, 3, 4, 5]
        >>> first(data, condition=lambda x: x % 2 == 0)
        2

    """
    if isinstance(data, set):
        raise TypeError(f'<data> expects a sequence. Got set. Use `one(data)` instead')

    return one(data, condition=condition)


def first_with_idx(data, *, condition=lambda _: True):
    """
    Returns the first element and its index from a sequence.

    Args:
        data (sequence): The sequence to search in.
        condition (function, optional): A function to specify the condition for comparison. Defaults to lambda _: True.

    Returns:
        tuple or None: A tuple containing the index and the first element
            that satisfies the condition specified by the condition function.

            When no item is found, returns None instead of (None, None).
            (None, None) lets you use unpacking everywhere, but it's a bad idea because
            it's evaluated as truthy in `if first_with_idx(...): ...`
            It would lead to a lot of counter-intuitive bugs, so it's better to avoid it

    Raises:
        TypeError: If data is a set. Use `one(data)` instead for sets.
        TypeError: If condition is not a callable function.

    Example:
        >>> first_with_idx([1, 2, 3, 4, 5])
        (0, 1)
        >>> first_with_idx([10, 20, 30, 40, 50], condition=lambda x: x % 3 == 0)
        (2, 30)
        >>> first_with_idx([1, 3, 5], condition=lambda x: x % 2 == 0)

    """
    if isinstance(data, set):
        raise TypeError(f'<data> expects a sequence. Got set. Use `one(data)` instead')
    elif not callable(condition):
        raise TypeError(f'<condition> expects a function. Got {type(condition).__name__}: {condition!r}')

    for i, v in enumerate(data):
        if condition(v):
            return (i, v)

    return None


def flatten(iterable):
    """
    Flatten one level of the iterable.

    Args:
        iterable (Iterable): The iterable to flatten.

    Returns:
        list: An iterable flattened by one level.

    Raises:
        TypeError: If the input is not an iterable.

    Examples:
        >>>

    """
    if not is_iterable(iterable):
        raise TypeError(f'<iterable> expects an iterable. Got {type(iterable).__name__}')

    flat = []
    for x in iterable:
        if is_iterable(x):
            flat.extend(x)
        else:
            flat.append(x)

    return flat


def flatten_recursive(iterable):
    """
    Recursively flattens an iterable.

    Args:
        iterable (Iterable): The iterable to be flattened.

    Returns:
        list: The flattened list.

    Example:
        >>> iterable = [1, [2, [3, 4], 5], 6]
        >>> flatten_recursive(iterable)
        [1, 2, 3, 4, 5, 6]

    """
    new = flatten(iterable)
    if new == iterable:
        return new
    else:
        return flatten_recursive(new)


def last(data, *, condition=lambda _: True):
    """
    Returns the last element from a sequence.

    Args:
        data (sequence): The sequence from which to retrieve the last element.
        condition (function, optional): A function to specify the condition for comparison. Defaults to lambda _: True.

    Returns:
        The last element from the sequence.

    Raises:
        TypeError: If data is a set. Use `one(data)` instead for sets.

    Example:
        >>> last([1, 2, 3, 4, 5])
        5

        >>> last([1, 2, 3, 4, 5], condition=lambda x: x % 2 == 0)
        4

    """
    if isinstance(data, set):
        raise TypeError(f'<data> expects a sequence. Got set. Use `one(data)` instead')

    if is_iterator(data) or isinstance(data, dict):
        data = list(data)

    return one(reversed(data), condition=condition)


def last_with_idx(data, *, condition=lambda _: True):
    """
    Returns the last element and its index from a sequence.

    Args:
        data (sequence): The sequence to search in.
        condition (function, optional): A function to specify the condition for comparison. Defaults to lambda _: True.

    Returns:
        tuple or None: A tuple containing the index and the first element
            that satisfies the condition specified by the condition function.

            When no item is found, returns None instead of (None, None).
            (None, None) lets you use unpacking everywhere, but it's a bad idea because
            it's evaluated as truthy in `if first_with_idx(...): ...`
            It would lead to a lot of counter-intuitive bugs, so it's better to avoid it

    Raises:
        TypeError: If data is a set. Use `one(data)` instead for sets.
        TypeError: If condition is not a callable function.

    Example:
        >>> last_with_idx([1, 2, 3, 4, 5])
        (4, 5)
        >>> last_with_idx([1, 2, 3, 4, 5], condition=lambda x: x % 2 == 0)
        (3, 4)
        >>> last_with_idx([1, 3, 5], condition=lambda x: x % 2 == 0)

    """
    if isinstance(data, set):
        raise TypeError(f'<data> expects a sequence. Got set. Use `one(data)` instead')
    elif not callable(condition):
        raise TypeError(f'<condition> expects a function. Got {type(condition).__name__}: {condition!r}')

    if is_iterator(data) or isinstance(data, dict):
        data = list(data)

    for i, v in enumerate(reversed(data), start=1):
        if condition(v):
            return (len(data) - i, v)

    return None


def map_dict_keys(f, data, *, condition=None):
    """
    Maps the keys of a dictionary to new keys using a function.

    Args:
        f (function): The function to apply to each key.
        data (dict): The dictionary to map the keys of.
        condition (function, optional): A condition function to selectively apply the mapping. Defaults to None.

    Returns:
        dict: A new dictionary with the mapped keys.

    Raises:
        TypeError: If f is not a callable function.
        TypeError: If data is not a dictionary.
        TypeError: If condition is neither None nor a callable function.

    Example:
        >>> def square(x):
        ...     return x ** 2
        ...
        >>> data = {1: 'one', 2: 'two', 3: 'three'}
        >>> map_dict_keys(square, data)
        {1: 'one', 4: 'two', 9: 'three'}

        >>> def is_even(x):
        ...     return x % 2 == 0
        ...
        >>> data = {1: 'one', 2: 'two', 3: 'three'}
        >>> map_dict_keys(square, data, condition=is_even)
        {1: 'one', 4: 'two', 3: 'three'}

    """
    if not callable(f):
        raise TypeError(f'<f> expects a function. Got {type(f).__name__}: {f!r}')
    elif not isinstance(data, dict):
        raise TypeError(f'<data> expects a dict. Got {type(data).__name__}: {data!r}')
    elif not(condition is None or callable(condition)):
        raise TypeError(f'<condition> expects None or a function. Got {type(condition).__name__}: {condition!r}')

    if condition is None:
        return {f(k): v for k, v in data.items()}
    else:
        return {f(k) if condition(k) else k: v for k, v in data.items()}


def map_dict_values(f, data, *, condition=None):
    """
    Maps the values of a dictionary to new values using a function.

    Args:
        f (function): The function to apply to each value.
        data (dict): The dictionary to map the values of.
        condition (function, optional): A condition function to selectively apply the mapping. Defaults to None.

    Returns:
        dict: A new dictionary with the mapped values.

    Raises:
        TypeError: If f is not a callable function.
        TypeError: If data is not a dictionary.
        TypeError: If condition is neither None nor a callable function.

    Example:
        >>> def uppercase(s):
        ...     return s.upper()
        ...
        >>> data = {'a': 'apple', 'b': 'banana', 'c': 'cherry'}
        >>> map_dict_values(uppercase, data)
        {'a': 'APPLE', 'b': 'BANANA', 'c': 'CHERRY'}

        >>> def is_long(s):
        ...     return len(s) > 5
        ...
        >>> data = {'a': 'apple', 'b': 'banana', 'c': 'cherry'}
        >>> map_dict_values(uppercase, data, condition=is_long)
        {'a': 'apple', 'b': 'BANANA', 'c': 'CHERRY'}

    """
    if not callable(f):
        raise TypeError(f'<f> expects a function. Got {type(f).__name__}: {f!r}')
    elif not isinstance(data, dict):
        raise TypeError(f'<data> expects a dict. Got {type(data).__name__}: {data!r}')
    elif not(condition is None or callable(condition)):
        raise TypeError(f'<condition> expects None or a function. Got {type(condition).__name__}: {condition!r}')

    if condition is None:
        return {k: f(v) for k, v in data.items()}
    else:
        return {k: f(v) if condition(v) else v for k, v in data.items()}


def map_recursive(f, data, *, condition=is_primitive, on_dict_keys=False):
    """
    Recursively maps a function over elements in a data structure, based on specified conditions.

    Args:
        f: The function to be applied to the elements. Must be callable.
        data: The data structure to be mapped over. Must be iterable.
        condition (optional):
            A function to determine whether an element should be mapped or not. If None, all elements
            are mapped. Default is is_primitive, which maps only primitives (not containers). (default: is_primitive)
        on_dict_keys (optional):
            Specifies whether to apply the mapping on dictionary keys.
            If True, the function will also be applied to dictionary keys. Default is False. (default: False)

    Returns:
        The data structure with the mapped elements.

    Raises:
        TypeError: If f is not callable, data is not iterable, condition is not None or callable,
         or on_dict_keys is not a boolean.

    Examples:
        1. Mapping a function over a list:
            >>> data = [1, 2, 3]
            >>> def square(x):
            ...     return x ** 2
            >>> map_recursive(square, data)
            [1, 4, 9]

        2. Mapping a function over a nested dictionary:
            >>> data = {'a': [1, 2, 3], 'b': {'c': [4, 5, 6]}}
            >>> def double(x):
            ...     return 2 * x
            >>> map_recursive(double, data)
            {'a': [2, 4, 6], 'b': {'c': [8, 10, 12]}}

    """
    if not callable(f):
        raise TypeError(f'<f> expects a function. Got {type(f).__name__}: {f!r}')
    elif not is_iterable(data):
        raise TypeError(f'<data> expects an iterable. Got {type(data).__name__}: {data!r}')
    elif not(condition is None or callable(condition)):
        raise TypeError(f'<condition> expects None or a function. Got {type(condition).__name__}: {condition!r}')
    elif not isinstance(on_dict_keys, bool):
        raise TypeError(f'<on_dict_keys> expects True or False. Got {type(on_dict_keys).__name__}: {on_dict_keys!r}')

    if is_iterator(data):
        new_data = (
            map_recursive(f, v, condition=condition, on_dict_keys=on_dict_keys)
            if is_iterable(v) else f(v) if condition(v) else v for v in data
        )
    elif isinstance(data, dict):
        new_data = {}
        for k, v in data.items():
            if on_dict_keys:
                if is_iterable(k):
                    k = map_recursive(f, k, condition=condition, on_dict_keys=on_dict_keys)
                if condition(k):
                    k = f(k)

            if is_iterable(v):
                v = map_recursive(f, v, condition=condition, on_dict_keys=on_dict_keys)
            if condition(v):
                v = f(v)

            new_data[k] = v
    else:
        new_data = type(data)(
            map_recursive(f, v, condition=condition, on_dict_keys=on_dict_keys)
            if is_iterable(v) else f(v) if condition(v) else v for v in data
        )

    if condition(new_data):
        return f(new_data)
    else:
        return new_data


def one(data, *, condition=lambda _: True):
    """
    Returns an element from an iterable that satisfies a given condition specified by the condition function.

    Args:
        data (iterable): The iterable to search in.
        condition (function, optional): A function to specify the condition. Defaults to lambda _: True.

    Returns:
        Any or None: An element that satisfies the condition. Returns None if no such element is found.

    Raises:
        TypeError: If condition is not a callable function.
        TypeError: If data is not an iterable.

    Example:
        >>> data = [1, 2, 3, 4, 5]
        >>> one(data)
        1

        >>> data = [1, 2, 3, 4, 5]
        >>> one(data, condition=lambda x: x % 2 == 0)
        2

        >>> data = [1, 3, 5]
        >>> one(data, condition=lambda x: x % 2 == 0)

    """
    if not callable(condition):
        raise TypeError(f'<condition> expects a function. Got {type(condition).__name__}: {condition!r}')
    elif not is_iterable(data):
        raise TypeError(f'<data> expects an iterable. Got {type(data).__name__}: {data!r}')

    try:
        return next(filter(condition, data))
    except StopIteration:
        return None


def pop_min(data, *, condition=lambda x: x, default=_UNDEFINED):
    """
    Removes and returns the minimum element from the collection.

    If <condition> is specified, it will be used to determine the minimum element

    Args:
        data: The collection from which the minimum element is to be removed. Must be iterable.
        condition (optional):
            A function to sort by. The min is determined by applying the function to each element.
            By default it uses the identity function. (default: lambda x: x)
        default (optional):
            A default value to return for empty collections. If not provided and the collection is empty,
            a ValueError is raised. (default: _UNDEFINED)

    Returns:
        A tuple containing the minimum element and the modified collection without the minimum element.

    Raises:
        TypeError: If data is not iterable, data is a dict, or condition is not callable.
        ValueError: If the collection is empty and a default value is not provided.

    Note:
        For dictionaries, use `pop_min_key` or `pop_min_value` instead.

    Examples:
        1. Pop the minimum element from a list:
            >>> data = [3, 1, 4, 1, 5, 9]
            >>> pop_min(data)
            (1, [3, 4, 1, 5, 9])

        2. Pop the minimum element from a list using a condition function:
            >>> data = ['apple', 'banana', 'cherry']
            >>> pop_min(data, condition=len)
            ('apple', ['banana', 'cherry'])

    """
    if not is_iterable(data):
        raise TypeError(f'<data> expects an iterable. Got {type(data).__name__}: {data!r}')
    elif isinstance(data, dict):
        raise TypeError(
            '<data> expects any iterable besides dict. '
            'For dict, use `pop_min_key` or `pop_min_value` instead')
    elif not callable(condition):
        raise TypeError(f'<condition> expects a function. Got {type(condition).__name__}: {condition!r}')

    original_type = type(data)
    data = [copy.deepcopy(item) for item in data]
    found = min(data, key=condition, default=_UNDEFINED)
    if found is _UNDEFINED and default is _UNDEFINED:
        raise ValueError('Cannot pop from an empty collection unless a default value is provided')

    data.remove(found)
    if original_type in (list, tuple, set, frozenset):
        return found, original_type(data)
    else:
        return found, data


def pop_max(data, *, condition=lambda x: x, default=_UNDEFINED):
    """
    Removes and returns the maximum element from the collection.

    If <condition> is specified, it will be used to determine the maximum element

    Args:
        data: The collection from which the maximum element is to be removed. Must be iterable.
        condition (optional):
            A function to sort by. The max is determined by applying the function to each element.
            By default it uses the identity function. (default: lambda x: x)
        default (optional):
            A default value to return for empty collections. If not provided and the collection is empty,
            a ValueError is raised. (default: _UNDEFINED)

    Returns:
        A tuple containing the maximum element and the modified collection without the maximum element.

    Raises:
        TypeError: If data is not iterable, data is a dict, or condition is not callable.
        ValueError: If the collection is empty and a default value is not provided.

    Note:
        For dictionaries, use `pop_max_key` or `pop_max_value` instead.

    Examples:
        1. Pop the maximum element from a list:
            >>> data = [3, 1, 4, 1, 5, 9]
            >>> pop_max(data)
            (9, [3, 1, 4, 1, 5])

        2. Pop the maximum element from a list using a condition function:
            >>> data = ['apple', 'banana', 'cherry']
            >>> pop_max(data, condition=len)
            ('banana', ['apple', 'cherry'])

    """
    if not is_iterable(data):
        raise TypeError(f'<data> expects an iterable. Got {type(data).__name__}: {data!r}')
    elif isinstance(data, dict):
        raise TypeError(
            '<data> expects any iterable besides dict. '
            'For dict, use `pop_max_key` or `pop_max_value` instead')
    elif not callable(condition):
        raise TypeError(f'<condition> expects a function. Got {type(condition).__name__}: {condition!r}')

    original_type = type(data)
    data = [copy.deepcopy(item) for item in data]
    found = max(data, key=condition, default=_UNDEFINED)
    if found is _UNDEFINED and default is _UNDEFINED:
        raise ValueError('Cannot pop from an empty collection unless a default value is provided')

    data.remove(found)
    if original_type in (list, tuple, set, frozenset):
        return found, original_type(data)
    else:
        return found, data


def pop_max_by_dict_key(data, *, condition=lambda x: x, default=_UNDEFINED):
    """
    Removes and returns the maximum (key, value) pair from the dictionary based on the dictionary key.

    If the <condition> argument is specified, it will be used to calculate the maximum.

    Args:
        data: The dictionary from which the maximum (key, value) pair is to be removed. Must be a dict.
        condition (optional):
            A function to sort by. The max (key, value) pair is determined by applying the function to each dict key.
            By default it uses the identity function. (default: lambda x: x)
        default (optional):
            A default value to return for empty collections. If not provided and the collection is empty,
            a ValueError is raised. (default: _UNDEFINED)

    Returns:
        A tuple containing the maximum (key, value) pair and the modified dictionary without the pair.

    Raises:
        TypeError: If data is not a dict or condition is not callable.
        ValueError: If the dictionary is empty and a default value is not provided.

    Examples:
        1. Pop the maximum (key, value) pair from a dictionary:
            >>> data = {'a': 1, 'b': 2, 'c': 3}
            >>> pop_max_by_dict_key(data)
            (('c', 3), {'a': 1, 'b': 2})

        2. Pop the maximum (key, value) pair from a dictionary using a condition function:
            >>> data = {'apple': 3, 'banana': 2, 'cherry': 5}
            >>> pop_max_by_dict_key(data, condition=len)
            (('banana', 2), {'apple': 3, 'cherry': 5})

    """
    if not isinstance(data, dict):
        raise TypeError(f'<data> expects a dict. Got {type(data).__name__}: {data!r}')
    elif not callable(condition):
        raise TypeError(f'<condition> expects a function. Got {type(condition).__name__}: {condition!r}')

    data = copy.deepcopy(data)
    dict_key = max(data.keys(), key=condition, default=_UNDEFINED)
    if dict_key is _UNDEFINED and default is _UNDEFINED:
        raise ValueError('Cannot pop from an empty dict unless a default value is provided')

    value = data.pop(dict_key)
    return (dict_key, value), data


def pop_min_by_dict_key(data, *, condition=lambda x: x, default=_UNDEFINED):
    """
    Removes and returns the minimum (key, value) pair from the dictionary based on the dictionary key.

    If <condition> is specified, it will be used to calculate the minimum.

    Args:
        data: The dictionary from which the minimum (key, value) pair is to be removed. Must be a dict.
        condition (optional):
            A function to sort by. The min (key, value) pair is determined by applying the function to each dict key.
            By default it uses the identity function. (default: lambda x: x)
        default (optional):
            A default value to return for empty collections. If not provided and the collection is empty,
            a ValueError is raised. (default: _UNDEFINED)

    Returns:
        A tuple containing the minimum (key, value) pair and the modified dictionary without the pair.

    Raises:
        TypeError: If data is not a dict or condition is not callable.
        ValueError: If the dictionary is empty and a default value is not provided.

    Examples:
        1. Pop the minimum (key, value) pair from a dictionary:
            >>> data = {'a': 1, 'b': 2, 'c': 3}
            >>> pop_min_by_dict_key(data)
            (('a', 1), {'b': 2, 'c': 3})

        2. Pop the minimum (key, value) pair from a dictionary using a condition function:
            >>> data = {'apple': 3, 'banana': 2, 'cherry': 5}
            >>> pop_min_by_dict_key(data, condition=len)
            (('apple', 3), {'banana': 2, 'cherry': 5})

    """
    if not isinstance(data, dict):
        raise TypeError(f'<data> expects a dict. Got {type(data).__name__}: {data!r}')
    elif not callable(condition):
        raise TypeError(f'<condition> expects a function. Got {type(condition).__name__}: {condition!r}')

    data = copy.deepcopy(data)
    dict_key = min(data.keys(), key=condition, default=_UNDEFINED)
    if dict_key is _UNDEFINED and default is _UNDEFINED:
        raise ValueError('Cannot pop from an empty dict unless a default value is provided')

    value = data.pop(dict_key)
    return (dict_key, value), data


def pop_max_by_dict_value(data, *, condition=lambda x: x, default=_UNDEFINED):
    """
    Removes and returns the maximum (key, value) pair from the dictionary based on the dictionary value.

    If <condition> is specified, it will be used to calculate the maximum.

    Args:
        data: The dictionary from which the maximum (key, value) pair is to be removed. Must be a dict.
        condition (optional):
            A function to sort by. The max (key, value) pair is determined by applying the function to each dict value.
            By default it uses the identity function. (default: lambda x: x)
        default (optional):
            A default value to return for empty collections. If not provided and the collection is empty,
            a ValueError is raised. (default: _UNDEFINED)

    Returns:
        A tuple containing the maximum (key, value) pair and the modified dictionary without the pair.

    Raises:
        TypeError: If data is not a dict or condition is not callable.
        ValueError: If the dictionary is empty and a default value is not provided.

    Examples:
        >>> data = {'a': 1, 'b': 2, 'c': 3}
        >>> pop_max_by_dict_value(data)
        (('c', 3), {'a': 1, 'b': 2})

    """
    if not isinstance(data, dict):
        raise TypeError(f'<data> expects a dict. Got {type(data).__name__}: {data!r}')
    elif not callable(condition):
        raise TypeError(f'<condition> expects a function. Got {type(condition).__name__}: {condition!r}')

    data = copy.deepcopy(data)
    item = max(data.items(), key=lambda kv: condition(kv[1]), default=_UNDEFINED)
    if item is _UNDEFINED and default is _UNDEFINED:
        raise ValueError('Cannot pop from an empty dict unless a default value is provided')

    _ = data.pop(item[0])
    return item, data


def pop_min_by_dict_value(data, *, condition=lambda x: x, default=_UNDEFINED):
    """
    Removes and returns the minimum (key, value) pair from the dictionary based on the dictionary value.

    If <condition> is specified, it will be used to calculate the minimum.

    Args:
        data: The dictionary from which the minimum (key, value) pair is to be removed. Must be a dict.
        condition (optional):
            A function to sort by. The min (key, value) pair is determined by applying the function to each dict value.
            By default it uses the identity function. (default: lambda x: x)
        default (optional):
            A default value to return for empty collections. If not provided and the collection is empty,
            a ValueError is raised. (default: _UNDEFINED)

    Returns:
        A tuple containing the minimum (key, value) pair and the modified dictionary without the pair.

    Raises:
        TypeError: If data is not a dict or condition is not callable.
        ValueError: If the dictionary is empty and a default value is not provided.

    Examples:
        >>> data = {'a': 1, 'b': 2, 'c': 3}
        >>> pop_min_by_dict_value(data)
        (('a', 1), {'b': 2, 'c': 3})

    """
    if not isinstance(data, dict):
        raise TypeError(f'<data> expects a dict. Got {type(data).__name__}: {data!r}')
    elif not callable(condition):
        raise TypeError(f'<condition> expects a function. Got {type(condition).__name__}: {condition!r}')

    data = copy.deepcopy(data)
    item = min(data.items(), key=lambda kv: condition(kv[1]), default=_UNDEFINED)
    if item is _UNDEFINED and default is _UNDEFINED:
        raise ValueError('Cannot pop from an empty dict unless a default value is provided')

    _ = data.pop(item[0])
    return item, data


def pick_all(needles, haystack):
    """
    Returns a list of elements from the haystack that match any of the given needles.

    Args:
        needles: The elements to search for in the haystack. Must be iterable.
        haystack: The collection to search within. Must be iterable.

    Returns:
        A list of elements from the haystack that match any of the needles.

    Raises:
        TypeError: If needles or haystack is not iterable.

    Examples:
        >>> needles = [2, 4, 6]
        >>> haystack = [1, 2, 3, 4, 5, 6]
        >>> pick_all(needles, haystack)
        [2, 4, 6]

    """
    if not is_iterable(needles):
        raise TypeError(f'<needles> expects an iterable. Got {type(needles).__name__}: {needles!r}')
    elif not is_iterable(haystack):
        raise TypeError(f'<haystack> expects an iterable. Got {type(haystack).__name__}: {haystack!r}')

    return [needle for needle in needles if is_any_of(needle, haystack)]


def pick_all_besides(needles, haystack):
    """
    Returns a list of elements from the needles that are not present in the haystack.

    Args:
        needles: The elements to search for in the haystack. Must be iterable.
        haystack: The collection to search within. Must be iterable.

    Returns:
        A list of elements from the needles that are not present in the haystack.

    Raises:
        TypeError: If needles or haystack is not iterable.

    Examples:
        >>> needles = [2, 4, 6, 7]
        >>> haystack = [1, 2, 3, 4, 5]
        >>> pick_all_besides(needles, haystack)
        [6, 7]

    """
    if not is_iterable(needles):
        raise TypeError(f'<needles> expects an iterable. Got {type(needles).__name__}: {needles!r}')
    elif not is_iterable(haystack):
        raise TypeError(f'<haystack> expects an iterable. Got {type(haystack).__name__}: {haystack!r}')

    return [needle for needle in needles if is_none_of(needle, haystack)]


def pick_any(needles, haystack):
    """
    Returns the first element from the needles that is present in the haystack.

    Args:
        needles: The elements to search for in the haystack. Must be iterable.
        haystack: The collection to search within. Must be iterable.

    Returns:
        The first element from the needles that is present in the haystack, or None if no matching element is found.

    Raises:
        TypeError: If needles or haystack is not iterable.

    Examples:
        >>> needles = [2, 4, 6]
        >>> haystack = [1, 2, 3, 4, 5, 6]
        >>> pick_any(needles, haystack)
        2

    """
    if not is_iterable(needles):
        raise TypeError(f'<needles> expects an iterable. Got {type(needles).__name__}: {needles!r}')
    elif not is_iterable(haystack):
        raise TypeError(f'<haystack> expects an iterable. Got {type(haystack).__name__}: {haystack!r}')

    for needle in needles:
        if is_any_of(needle, haystack):
            return needle

    return None


def sort_dict_by_keys(data, *, condition=None, reverse=False):
    """
    Sorts a dictionary by its keys and returns a new dictionary.

    Args:
        data: The dictionary to be sorted. Must be a dict.
        condition (optional):
            A function to sort by. The keys are sorted based on the result of applying this function to each key.
            If None, the keys are sorted naturally. (default: None)
        reverse (optional):
            Reverses the sorting order. Sort in descending order if True and ascending otherwise. (default: False)

    Returns:
        A new dictionary sorted by keys.

    Raises:
        TypeError: If data is not a dict, condition is not callable or None, or reverse is not a boolean.

    Examples:
        >>> data = {'b': 2, 'a': 1, 'c': 3}
        >>> sort_dict_by_keys(data)
        {'a': 1, 'b': 2, 'c': 3}

    """
    if not isinstance(data, dict):
        raise TypeError(f'<data> expects a dict. Got {type(data).__name__}: {data!r}')
    elif not(condition is None or callable(condition)):
        raise TypeError(f'<condition> expects None or a function. Got {type(condition).__name__}: {condition!r}')
    elif not isinstance(reverse, bool):
        raise TypeError(f'<reverse> expects True or False. Got {type(reverse).__name__}: {reverse!r}')

    if condition is None:
        return dict(sorted(data.items(), reverse=reverse))
    else:
        return dict(sorted(data.items(), key=lambda kv: condition(kv[0]), reverse=reverse))


def sort_dict_by_keys_recursive(data, *, condition=None, reverse=False):
    """
    Recursively sorts a dictionary and its nested containers by their keys and returns a new structure.

    Args:
        data: The data structure to be sorted. Must be an iterable.
        condition (optional):
            A function to sort by. The keys are sorted based on the result of applying this function to each key.
            If None, the keys are sorted naturally. (default: None)
        reverse (optional):
            Reverses the sorting order. Sort in descending order if True and ascending otherwise. (default: False)

    Returns:
        A new structure with the nested dictionaries and containers sorted by keys.

    Raises:
        TypeError: If data is not an iterable, condition is not callable or None, or reverse is not a boolean.

    Examples:
        >>> data = {'b': {'c': 30, 'b': 20, 'a': 10}, 'a': {'c': 3, 'b': 2, 'a': 1}}
        >>> sort_dict_by_keys_recursive(data)
        {'a': {'a': 1, 'b': 2, 'c': 3}, 'b': {'a': 10, 'b': 20, 'c': 30}}

    """
    if not is_iterable(data):
        raise TypeError(f'<data> expects an iterable. Got {type(data).__name__}: {data!r}')
    elif not(condition is None or callable(condition)):
        raise TypeError(f'<condition> expects None or a function. Got {type(condition).__name__}: {condition!r}')
    elif not isinstance(reverse, bool):
        raise TypeError(f'<reverse> expects True or False. Got {type(reverse).__name__}: {reverse!r}')

    if isinstance(data, dict):
        return sort_dict_by_keys(
            {k: sort_dict_by_keys_recursive(v, condition=condition, reverse=reverse) if is_iterable(v) else v
                for k, v in data.items()},
            condition=condition,
            reverse=reverse)
    elif is_iterator(data):
        return (sort_dict_by_keys_recursive(v, condition=condition, reverse=reverse) if is_iterable(v) else v
                for v in data)
    elif is_iterable(data):
        return type(data)(sort_dict_by_keys_recursive(v, condition=condition, reverse=reverse) if is_iterable(v) else v
                          for v in data)


def sort_dict_by_values(data, *, condition=None, reverse=False):
    """
    Sorts a dictionary by its values and returns a new dictionary.

    Args:
        data: The dictionary to be sorted. Must be a dict.
        condition (optional):
            A function to sort by. The keys are sorted based on the result of applying this function to each value.
            If None, the values are sorted naturally. (default: None)
        reverse (optional):
            Reverses the sorting order. Sort in descending order if True and ascending otherwise. (default: False)

    Returns:
        A new dictionary sorted by values.

    Raises:
        TypeError: If data is not a dict, condition is not callable or None, or reverse is not a boolean.

    Examples:
        >>> data = {'a': 3, 'b': 1, 'c': 2}
        >>> sort_dict_by_values(data)
        {'b': 1, 'c': 2, 'a': 3}

    """
    if not isinstance(data, dict):
        raise TypeError(f'<data> expects a dict. Got {type(data).__name__}: {data!r}')
    elif not(condition is None or callable(condition)):
        raise TypeError(f'<condition> expects None or a function. Got {type(condition).__name__}: {condition!r}')
    elif not isinstance(reverse, bool):
        raise TypeError(f'<reverse> expects True or False. Got {type(reverse).__name__}: {reverse!r}')

    if condition is None:
        return dict(sorted(data.items(), key=lambda kv: kv[1], reverse=reverse))
    else:
        return dict(sorted(data.items(), key=lambda kv: condition(kv[1]), reverse=reverse))


def sort_dict_by_values_recursive(data, *, condition=None, reverse=False):
    """
    Recursively sorts a dictionary and its nested containers by their values and returns a new structure.

    Args:
        data: The data structure to be sorted. Must be an iterable.
        condition (optional):
            A function to sort by. The keys are sorted based on the result of applying this function to each value.
            If None, the values are sorted naturally. (default: None)
        reverse (optional):
            Reverses the sorting order. Sort in descending order if True and ascending otherwise. (default: False)

    Returns:
        A new structure with the nested dictionaries and containers sorted by values.

    Raises:
        TypeError: If data is not an iterable, condition is not callable or None, or reverse is not a boolean.

    """
    if not is_iterable(data):
        raise TypeError(f'<data> expects an iterable. Got {type(data).__name__}: {data!r}')
    elif not(condition is None or callable(condition)):
        raise TypeError(f'<condition> expects None or a function. Got {type(condition).__name__}: {condition!r}')
    elif not isinstance(reverse, bool):
        raise TypeError(f'<reverse> expects True or False. Got {type(reverse).__name__}: {reverse!r}')

    if isinstance(data, dict):
        return sort_dict_by_values(
            {k: sort_dict_by_values_recursive(v, condition=condition, reverse=reverse) if is_iterable(v) else v
                for k, v in data.items()},
            condition=condition,
            reverse=reverse)
    elif is_iterator(data):
        return (sort_dict_by_values_recursive(v, condition=condition, reverse=reverse) if is_iterable(v) else v
                for v in data)
    elif is_iterable(data):
        return type(data)(
            sort_dict_by_values_recursive(v, condition=condition, reverse=reverse) if is_iterable(v) else v
            for v in data)


def sort_recursive(
    data, *, condition=lambda x: x, reverse=False, sort_dicts_by='keys',
    sort_iters=True, sort_lists=True, sort_sets=True, sort_tuples=True
):
    """
    Recursively sorts an iterable data structure with customizable sorting options.
    The collections are sorted from the innermost to the outermost one

    Args:
        data: The data structure to be sorted. Must be an iterable.
        condition (optional):
            A function to sort by. Sort based on the result of applying this function to each element.
            By default it uses the identity function. (default: lambda x: x)
        reverse (optional):
            Reverses the sorting order. Sort in descending order if True and ascending otherwise. (default: False)
        sort_dicts_by (optional):
            Specifies how to sort dicts. Sort by 'keys' or 'values'. If None, don't sort dicts. (default: 'keys')
        sort_iters (optional):
            Enable sort for iterable types (e.g., generator, range). default: True)
        sort_lists (optional): Enable sort for lists.(default: True)
        sort_sets (optional): Enable sort for sets. (default: True)
        sort_tuples (optional): Enable sort for tuples. (default: True)

    Returns:
        A new sorted data structure with the same nested structure.

    Raises:
        TypeError: If data is not an iterable or condition is not callable.
        ValueError: If sort_dicts_by has an invalid value.

    Examples:
        >>> data = [{'c': 3, 'b': {'z': 3, 'x': 2, 'y': 1}, 'a': 1}]
        >>> sort_recursive(data)
        [{'a': 1, 'b': {'x': 2, 'y': 1, 'z': 3}, 'c': 3}]

    """
    if not is_iterable(data):
        raise TypeError(f'<data> expects an iterable. Got {type(data).__name__}')
    elif not callable(condition):
        raise TypeError(f'<condition> expects a function. Got {type(condition).__name__}')
    elif not isinstance(reverse, bool):
        raise TypeError(f'<reverse> expects True or False. Got {type(reverse).__name__}')
    elif is_none_of(sort_dicts_by, ('keys', 'values', None)):
        raise ValueError(f'<sort_dicts_by> expects "keys", "values" or None. Got {sort_dicts_by!r}')
    elif not isinstance(sort_iters, bool):
        raise TypeError(f'<sort_iters> expects True or False. Got {type(sort_iters).__name__}')
    elif not isinstance(sort_lists, bool):
        raise TypeError(f'<sort_lists> expects True or False. Got {type(sort_lists).__name__}')
    elif not isinstance(sort_sets, bool):
        raise TypeError(f'<sort_sets> expects True or False. Got {type(sort_sets).__name__}')
    elif not isinstance(sort_tuples, bool):
        raise TypeError(f'<sort_tuples> expects True or False. Got {type(sort_tuples).__name__}')

    initial_type = type(data)
    kwargs = {
        'condition': condition,
        'reverse': reverse,
        'sort_dicts_by': sort_dicts_by,
        'sort_iters': sort_iters,
        'sort_lists': sort_lists,
        'sort_sets': sort_sets,
        'sort_tuples': sort_tuples,
    }

    if isinstance(data, dict):
        # You shouldn't sort the keys, even if they are tuple,
        # as they are likely to be used as IDs
        inner_sorted = {k: sort_recursive(v, **kwargs) if is_iterable(v) else v for k, v in data.items()}
        if sort_dicts_by == 'keys':
            return sort_dict_by_keys(inner_sorted, condition=condition, reverse=reverse)
        elif sort_dicts_by == 'values':
            return sort_dict_by_values(inner_sorted, condition=condition, reverse=reverse)
        else:
            return inner_sorted

    inner_sorted = (sort_recursive(v, **kwargs) if is_iterable(v) else v for v in data)
    if is_iterator(data) and not sort_iters:
        return inner_sorted
    elif isinstance(data, list) and not sort_lists:
        return list(inner_sorted)
    elif isinstance(data, set) and not sort_sets:
        return set(inner_sorted)
    elif isinstance(data, tuple) and not sort_tuples:
        return tuple(inner_sorted)
    elif is_iterator(data) or isinstance(data, (set, list, tuple)):
        data = sorted(inner_sorted, key=condition, reverse=reverse)
        if initial_type == tuple:
            return tuple(data)
        else:
            return data
    else:  # pragma: no cover
        raise AssertionError(f'Unexpected branch for <data> of type {type(data).__name__}')


def swap_keys_and_values(data):
    """
    Swaps the keys and values in a dictionary.

    Args:
        data: The dictionary to swap keys and values. Must be a dict.

    Returns:
        A new dictionary with keys and values swapped.

    Raises:
        TypeError: If data is not a dict.

    Examples:
        Swap keys and values in a dictionary:
            >>> swap_keys_and_values({'a': 1, 'b': 2, 'c': 3})
            {1: 'a', 2: 'b', 3: 'c'}

    """
    if not isinstance(data, dict):
        raise TypeError(f'<data> expects a dict. Got {type(data).__name__}')

    return {v: k for k, v in data.items()}


def xor(*values, condition=bool):
    """
    Performs an exclusive OR (XOR) operation on multiple values using a condition function.

    Args:
        *values: Multiple values to perform the XOR operation on. At least two values are required.
        condition (optional):
            A function to determine the truthiness of each value. (default: bool)

    Returns: The value satisfying the XOR condition or None if the condition is False.

    Raises:
        ValueError: If less than two values are provided.
        TypeError: If condition is not callable.

    Examples:
        1. Perform XOR operation on a list of integers:
            >>> xor(0, 1, 0, 1, 0)

            >>> xor(0, 0, 0, 1, 0)
            1

        2. Perform XOR operation on a list of strings using a condition function:
            >>> xor('hello', '', 'world', condition=len)

            >>> xor('', 'hello', '', condition=len)
            'hello'

    """
    if len(values) < 2:
        raise ValueError('<values> expects at least two values')
    elif not callable(condition):
        raise TypeError(f'<condition> expects a function. Got {type(condition).__name__}')

    v = None
    for value in values:
        if condition(value):
            if v is None:
                v = value
            else:
                return None

    return v


def xor_with_idx(*values, condition=bool):
    """
    Performs a XOR (exclusive OR) operation on multiple values using a condition function.

    Args:
        *values: Multiple values to perform the XOR operation on. At least two values are required.
        condition (optional):
            A function to determine the truthiness of each value. (default: bool)

    Returns: The (index, value) satisfying the XOR condition or None if the condition is False.

    Raises:
        ValueError: If less than two values are provided.
        TypeError: If condition is not callable.

    Examples:
        1. Perform XOR operation on a list of integers:
            >>> xor_with_idx(0, 1, 0, 1, 0)

            >>> xor_with_idx(0, 0, 0, 1, 0)
            (3, 1)

        2. Perform XOR operation on a list of strings using a condition function:
            >>> xor_with_idx('hello', '', 'world', condition=len)

            >>> xor_with_idx('', 'hello', '', condition=len)
            (1, 'hello')

    """
    if len(values) < 2:
        raise ValueError('<values> expects at least two values')
    elif not callable(condition):
        raise TypeError(f'<condition> expects a function. Got {type(condition).__name__}')

    v = None
    for i, value in enumerate(values):
        if condition(value):
            if v is None:
                v = (i, value)
            else:
                return None

    return v


if __name__ == '__main__':  # pragma: no cover
    import doctest
    doctest.testmod()
